import os
import subprocess
from configparser import ConfigParser

import boto3
import pytest
from moto import mock_s3

from redun import File, Scheduler, task
from redun.executors.aws_batch import AWSBatchExecutor
from redun.expression import TaskExpression
from redun.file import Dir
from redun.scripting import (
    ScriptError,
    exec_script,
    get_command_eof,
    get_wrapped_command,
    prepare_command,
    script,
)
from redun.tests.utils import use_tempdir


@use_tempdir
def test_redirect() -> None:
    """
    Shell redirection should use tee to create files for stderr and stdout.

    This test documents the technique works.
    """
    proc = subprocess.run(
        [
            "bash",
            "-c",
            """
            (echo hello) 2> >(tee >(cat > stderr) >&2) | tee >(cat > stdout)
            """,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    assert File("stdout").read() == "hello\n"
    assert File("stderr").read() == ""
    assert proc.stdout == b"hello\n"
    assert proc.stderr == b""
    assert proc.returncode == 0

    proc = subprocess.run(
        [
            "bash",
            "-c",
            "-o",
            "pipefail",
            "(bad_command) 2> >(tee >(cat > stderr) >&2) | "
            "tee >(cat > stdout) || (echo fail; exit 1)",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    assert File("stdout").read() == ""
    assert "command not found" in File("stderr").read()
    assert proc.stdout == b"fail\n"
    assert b"command not found" in proc.stderr
    assert proc.returncode != 0


def test_prepare_command() -> None:
    """
    Commands should be dedented.
    """
    assert (
        prepare_command(
            """
        #!/bin/bash
        echo hello
        """
        )
        == """\
#!/bin/bash
echo hello"""
    )


def test_prepare_command_default_interpreter() -> None:
    """
    Commands should be dedented.
    """
    assert (
        prepare_command(
            """
        echo hello
        """
        )
        == """\
#!/usr/bin/env bash
set -exo pipefail
echo hello"""
    )


def test_exec_script() -> None:
    """
    Execute a python script.
    """
    assert (
        exec_script(
            """\
#!/usr/bin/env python
print('hello')
"""
        )
        == b"hello\n"
    )


def test_script_task(scheduler: Scheduler) -> None:
    """
    Tasks should be definable as shell scripts.
    """

    @task(script=True)
    def task1(message):
        return """echo Hello, {message}!""".format(message=message)

    assert scheduler.run(task1("World")) == b"Hello, World!\n"


def test_script_task_error(scheduler: Scheduler) -> None:
    """
    Script tasks should raise errors from shell script.
    """

    @task(script=True)
    def task1():
        return "bad_command"

    @task(script=True)
    def task2():
        # Multi-line script should exit on first error by default.
        return """
bad_command
echo hello
        """

    with pytest.raises(ScriptError):
        scheduler.run(task1())

    with pytest.raises(ScriptError):
        scheduler.run(task2())


def test_python_script_task(scheduler: Scheduler) -> None:
    """
    Script tasks should be able to use custom interpreters.
    """

    @task(script=True)
    def task1(message):
        return """
        #!/usr/bin/env python
        print('Hello, {message}!')
        """.format(
            message=message
        )

    assert scheduler.run(task1("World")) == b"Hello, World!\n"


def test_python_script_task2(scheduler: Scheduler) -> None:
    """
    script() should use custom interpreters.
    """
    result = script(
        """
        #!/usr/bin/env python
        print('Hello, World!')
        """,
        executor="default",
    )
    assert isinstance(result, TaskExpression)
    assert scheduler.run(result) == b"Hello, World!\n"


def test_default_shell(scheduler: Scheduler) -> None:
    """
    script() should use bash as default interpreter.
    """
    result = script(
        """
        # Use a bash only syntax.
        cat <(echo ok)
        """
    )
    assert scheduler.run(result) == b"ok\n"


def test_script_error(scheduler: Scheduler) -> None:
    """
    Scripts should propagate their errors.
    """

    @task()
    def task1():
        return script(
            """
            echo message > /dev/stderr
            bad_prog 1 2 3
            """
        )

    with pytest.raises(ScriptError) as error:
        scheduler.run(task1())

    assert "message" in error.value.message
    assert "bad_prog: command not found" in error.value.message


def test_script_outputs(scheduler: Scheduler) -> None:
    """
    script() should be able to define an output structure.
    """
    result = script(
        """
        #!/bin/sh
        echo 'Hello, World!'
        """,
        executor="default",
        outputs={"my_output": 10, "stdout": File("-")},
    )
    assert isinstance(result, TaskExpression)
    assert scheduler.run(result) == {
        "my_output": 10,
        "stdout": b"Hello, World!\n",
    }


@use_tempdir
def test_script_file(scheduler: Scheduler) -> None:
    result = script(
        """
        #!/bin/sh
        echo 'hello' > hello.txt
        echo 'good bye' > bye.txt
        """,
        outputs={"hello": File("hello.txt"), "bye": File("bye.txt")},
    )

    result = scheduler.run(result)
    assert result["hello"].read() == "hello\n"
    assert result["bye"].read() == "good bye\n"


def test_command_eof() -> None:
    command = """
run-prog --x 10
ls my-dir
"""
    assert get_command_eof(command) == "EOF"

    command = """
run-prog --x 10 <<"EOF"
    ls my-dir
EOF
"""
    assert get_command_eof(command) == "EOF1"

    command = """
run-prog1 <<"EOF"
run-prog2 --x 10 <<"EOF1"
    ls my-dir
EOF1
EOF
"""
    assert get_command_eof(command) == "EOF2"


def test_wrapped_command() -> None:
    command = """\
#!/bin/bash
echo hello
"""
    assert "EOF" in get_wrapped_command(command)


def test_exec_wrapped_command() -> None:
    command = """\
#!/bin/bash
echo hello
"""
    wrapped_command = get_wrapped_command(command)
    assert subprocess.check_output(wrapped_command, shell=True) == b"hello\n"


def test_script_tempdir(scheduler: Scheduler) -> None:
    result = script(
        """
        #!/bin/sh
        echo 'hello' > hello.txt
        echo 'good bye' > bye.txt
        """,
        tempdir=True,
        outputs={"hello": File("hello.txt"), "bye": File("bye.txt")},
    )

    result = scheduler.run(result)

    # tempdir has been cleaned up.
    assert not result["hello"].exists()
    assert not result["bye"].exists()


@use_tempdir
def test_script_staging(scheduler: Scheduler) -> None:
    basedir = os.getcwd()

    hello_path = os.path.join(basedir, "remote_hello.txt")
    bye_path = os.path.join(basedir, "remote_bye.txt")

    result = script(
        """
        #!/bin/sh
        echo 'hello' > hello.txt
        echo 'good bye' > bye.txt
        """,
        tempdir=True,
        outputs={
            "hello": File(hello_path).stage("hello.txt"),
            "bye": File(bye_path).stage("bye.txt"),
        },
    )

    result = scheduler.run(result)

    # Remote files should still exist.
    assert result["hello"].exists()
    assert result["bye"].exists()
    assert result["hello"].path == hello_path
    assert result["bye"].path == bye_path
    assert result["hello"].read() == "hello\n"
    assert result["bye"].read() == "good bye\n"


@use_tempdir
def test_script_staging_dir(scheduler: Scheduler) -> None:
    basedir = os.getcwd()

    remote_dir = os.path.join(basedir, "remote")

    File(remote_dir + "/in/a.txt").write("a")
    File(remote_dir + "/in/b.txt").write("b")
    File(remote_dir + "/in/c/d.txt").write("d")

    result = script(
        """
        #!/bin/sh
        mkdir -p out
        cat in/a.txt in/b.txt in/c/d.txt > out/z
        echo 'hello' > out/y
        """,
        tempdir=True,
        inputs={Dir(remote_dir + "/in").stage("in")},
        outputs={"out": Dir(remote_dir + "/out").stage("out")},
    )

    result = scheduler.run(result)

    # Remote files should still exist.
    assert result["out"].exists()
    assert result["out"].file("z").read() == "abd"
    assert result["out"].file("y").read() == "hello\n"


@use_tempdir
def test_script_invalid(scheduler: Scheduler) -> None:
    """
    script() should be reactive to invalidated output.
    """

    expr = script(
        """
        echo hi > local
        """,
        outputs=[File("remote").stage("local")],
    )

    # Run the workflow once.
    [out_file] = scheduler.run(expr)
    assert out_file.read() == "hi\n"

    # Invalidate the output file by overwriting it.
    File("remote").write("bye")

    # Rerunning the workflow should reproduce the same output.
    [out_file] = scheduler.run(expr)
    assert out_file.read() == "hi\n"

    # Invalidate the output file by deleting.
    File("remote").remove()

    # Rerunning the workflow should reproduce the same output.
    scheduler.run(expr)
    assert File("remote").read() == "hi\n"


@use_tempdir
def test_script_staging_input_change(scheduler: Scheduler) -> None:
    """
    script() should be reactive to changing inputs.
    """

    File("input_remote").write("hello")

    expr = script(
        """
        cat input_local > output_local
        """,
        inputs=[File("input_remote").stage("input_local")],
        outputs=File("output_remote").stage("output_local"),
    )
    assert scheduler.run(expr).read() == "hello"

    # Change input.
    File("input_remote").write("hello2")

    expr = script(
        """
        cat input_local > output_local
        """,
        inputs=[File("input_remote").stage("input_local")],
        outputs=File("output_remote").stage("output_local"),
    )
    assert scheduler.run(expr).read() == "hello2"


@mock_s3
def _test_script_staging_s3(scheduler: Scheduler) -> None:
    s3_client = boto3.client("s3", region_name="us-east-1")
    s3_client.create_bucket(Bucket="example-bucket")

    hello_path = "s3://example-bucket/_hello.txt"
    bye_path = "s3://example-bucket/bye.txt"

    result = script(
        """
        #!/bin/sh
        echo 'hello' > hello.txt
        echo 'good bye' > bye.txt
        """,
        tempdir=True,
        outputs={
            "hello": File(hello_path).stage("hello.txt"),
            "bye": File(bye_path).stage("bye.txt"),
        },
    )

    result = scheduler.run(result)

    # Remote files should still exist.
    assert result["hello"].exists()
    assert result["bye"].exists()
    assert result["hello"].path == hello_path
    assert result["bye"].path == bye_path
    assert result["hello"].read() == "hello\n"
    assert result["bye"].read() == "good bye\n"


@mock_s3
def _test_script_task_aws_batch():
    @task(executor="batch", script=True)
    def task1(message):
        return """echo Hello, {message}!""".format(message=message)

    s3_client = boto3.client("s3", region_name="us-east-1")
    s3_client.create_bucket(Bucket="example-bucket")

    config = ConfigParser()
    config.read_dict(
        {
            "batch": {
                "image": "ubuntu",
                "queue": "queue",
                "s3_scratch": "s3://example-bucket/redun",
                "debug": True,
            }
        }
    )

    executor = AWSBatchExecutor("batch", None, config=config["batch"])
    scheduler = Scheduler()
    scheduler.executors["batch"] = executor
    executor.scheduler = scheduler
    assert scheduler.run(task1, ["World"]) == b"Hello, World!\n"



from itertools import chain
from typing import Any

import pytest
from sqlalchemy import Column, Integer, String, create_engine
from sqlalchemy.orm import Session, declarative_base, sessionmaker

from redun.db_utils import filter_in, get_or_create, query_filter_in

Base: Any = declarative_base()


class Person(Base):
    __tablename__ = "person"

    id = Column(Integer, primary_key=True)
    name = Column(String)


@pytest.fixture
def session():
    """
    Fixture for test database.
    """

    # Initialize in memory database.
    engine = create_engine("sqlite:///:memory:")
    Session = sessionmaker(bind=engine)
    session = Session()
    Base.metadata.create_all(bind=engine)

    # Create some example rows.
    session.add(Person(name="Alice"))
    session.add(Person(name="Bob"))
    session.add(Person(name="Claire"))
    session.add(Person(name="Dan"))
    session.commit()

    return session


def test_get_or_create(session: Session) -> None:
    """
    get_or_create should ensure records exist and update them if needed.
    """

    # Assert db initial state.
    assert session.query(Person.name).order_by(Person.name).all() == [
        ("Alice",),
        ("Bob",),
        ("Claire",),
        ("Dan",),
    ]

    # New row.
    row, created = get_or_create(session, Person, {"name": "Eve"})
    assert row.name == "Eve"
    assert created

    # Existing row.
    row, created = get_or_create(session, Person, {"name": "Alice"})
    assert row.name == "Alice"
    assert not created

    # Update existing row.
    row, created = get_or_create(session, Person, {"name": "Bob"}, update={"name": "Bobby"})
    assert row.name == "Bobby"
    assert not created

    # Update non-existent row.
    row, created = get_or_create(session, Person, {"name": "Fred"}, update={"name": "Freddie"})
    assert row.name == "Freddie"
    assert created


def test_filter_in(session: Session) -> None:
    """
    filter_in should allow large filter-IN clauses for sqlalchemy.
    """

    # Assert db initial state.
    assert session.query(Person.name).order_by(Person.name).all() == [
        ("Alice",),
        ("Bob",),
        ("Claire",),
        ("Dan",),
    ]

    query = session.query(Person.name).order_by(Person.name)

    rows = list(filter_in(query, Person.name, ["Alice", "Bob", "Eve"], chunk=1))
    assert rows == [("Alice",), ("Bob",)]

    queries = query_filter_in(query, Person.name, ["Alice", "Bob", "Eve"], chunk=1)
    rows = list(chain(*(subquery.all() for subquery in queries)))
    assert rows == [("Alice",), ("Bob",)]



import dataclasses
import os
import time
from traceback import FrameSummary
from typing import Any, Dict, List, Optional, Sequence, Tuple
from unittest.mock import Mock, patch

import pytest
from sqlalchemy.orm import Session

from redun import Scheduler, task
from redun.backends.db import Execution
from redun.backends.db import Job as JobDb
from redun.backends.db import RedunBackendDb
from redun.config import Config
from redun.expression import SchedulerExpression
from redun.promise import Promise
from redun.scheduler import (
    DryRunResult,
    Frame,
    Job,
    Task,
    Traceback,
    catch,
    catch_all,
    cond,
    scheduler_task,
)
from redun.task import CacheScope, PartialTask, SchedulerTask
from redun.tests.utils import assert_match_lines, use_tempdir
from redun.utils import map_nested_value
from redun.value import Value, get_type_registry


@dataclasses.dataclass
class Node:
    value: str
    next_: Optional["Node"]


def test_simple(scheduler: Scheduler) -> None:
    """
    A single task should execute.
    """

    @task()
    def workflow():
        return "ok"

    assert scheduler.run(workflow()) == "ok"


def test_call(scheduler: Scheduler) -> None:
    """
    A subtask call should evaluate.
    """

    @task()
    def task1(x):
        return "hello {}".format(x)

    @task()
    def workflow():
        result = task1("world")
        return result

    assert scheduler.run(workflow()) == "hello world"


def test_task_args(scheduler: Scheduler) -> None:
    """
    A task result can be used as an argument to another task.
    """

    @task()
    def task1():
        return "world"

    @task()
    def task2(x):
        return "hello {}".format(x)

    @task()
    def workflow():
        result = task1()
        result2 = task2(result)
        return result2

    assert scheduler.run(workflow()) == "hello world"


def test_call_deep(scheduler: Scheduler) -> None:
    @task()
    def task1(a):
        return "task1({})".format(a)

    @task()
    def combine(*values):
        return ",".join(values)

    @task()
    def task2(a_list):
        results = [task1(a) for a in a_list]
        return combine(*results)

    @task()
    def workflow():
        values = ["alice", "bob", "claire"]
        result = task2(values)
        return result

    assert scheduler.run(workflow()) == "task1(alice),task1(bob),task1(claire)"


@use_tempdir
def test_scheduler_backend() -> None:
    """
    Backend should set its db_uri correctly.
    """
    # Default scheduler should us in memory backend db.
    scheduler = Scheduler()
    assert isinstance(scheduler.backend, RedunBackendDb)
    assert scheduler.backend.db_uri == "sqlite:///:memory:"

    # If config specifies a db_uri, we should use it.
    os.makedirs(".redun")
    scheduler = Scheduler(config=Config({"backend": {"db_uri": "sqlite:///.redun/redun.db"}}))
    scheduler.load()
    assert isinstance(scheduler.backend, RedunBackendDb)
    assert scheduler.backend.db_uri == "sqlite:///.redun/redun.db"
    assert os.path.exists(".redun/redun.db")


def test_scheduler_incremental(scheduler: Scheduler) -> None:
    """
    Scheduler should incrementally execute tasks that change their hash.
    """
    task_calls = []

    @task()
    def run_bwa(reads):
        task_calls.append("run_bwa")
        align = "align({})".format(reads)
        return align

    @task()
    def run_gatk(align):
        task_calls.append("run_gatk")
        calls = "calls({})".format(align)
        return calls

    @task()
    def workflow():
        task_calls.append("workflow")
        reads = "my reads"
        align = run_bwa(reads)
        calls = run_gatk(align)
        return calls

    # Scheduler should call all the tasks and return the correct result.
    result = scheduler.run(workflow())

    assert result == "calls(align(my reads))"
    assert task_calls == ["workflow", "run_bwa", "run_gatk"]

    # Running the workflow again should not trigger any more task calls.
    task_calls = []
    result = scheduler.run(workflow())
    assert result == "calls(align(my reads))"
    assert task_calls == []

    # Simulate a code change to run_gatk task.

    @task()  # type: ignore[no-redef]
    def run_gatk(align):
        task_calls.append("run_gatk")
        calls = "calls2({})".format(align)
        return calls

    task_calls = []
    result = scheduler.run(workflow())

    assert result == "calls2(align(my reads))"
    assert task_calls == ["run_gatk"]

    # A reverted code change should be fully memoized too.

    @task()  # type: ignore[no-redef]
    def run_gatk(align):
        task_calls.append("run_gatk")
        calls = "calls({})".format(align)
        return calls

    task_calls = []
    result = scheduler.run(workflow())

    assert result == "calls(align(my reads))"
    assert task_calls == []


def test_scheduler_multiple_calls(scheduler: Scheduler) -> None:
    """
    We need to memoize the same function with different arguments.
    """

    @task()
    def task1(arg):
        return "hello {}".format(arg)

    @task()
    def task2():
        return [task1("alice"), task1("bob")]

    # Scheduler should call all the tasks and return the correct result.
    assert scheduler.run(task2()) == ["hello alice", "hello bob"]


def test_scheduler_recursive(scheduler: Scheduler) -> None:
    """
    Scheduler should handle recursive task calls.
    """

    task_calls = []

    @task()
    def add(a, b):
        return a + b

    @task()
    def fib(n):
        task_calls.append(n)
        if n <= 0:
            return 0
        elif n <= 2:
            return 1
        else:
            return add(fib(n - 1), fib(n - 2))

    # Scheduler should call all the tasks and return the correct result.
    assert scheduler.run(fib(1)) == 1
    assert scheduler.run(fib(2)) == 1
    assert scheduler.run(fib(3)) == 2
    assert scheduler.run(fib(5)) == 5
    assert len(task_calls) == 5


def test_destructure(scheduler: Scheduler):
    """
    Workflows should be able to use Expression destructuring.
    """

    @task()
    def task1():
        return {
            "output1": 10,
            "output2": 20,
        }

    @task()
    def task2(x):
        return x + 1

    @task()
    def workflow():
        outputs = task1()
        a = task2(outputs["output1"])
        b = task2(outputs["output2"])
        return [a, b]

    assert scheduler.run(workflow()) == [11, 21]


def test_return_tuple_type(scheduler: Scheduler) -> None:
    """
    Tasks with return type Tuple should support destructuring.
    """

    @task()
    def make_pair() -> Tuple[str, int]:
        return ("Bob", 20)

    @task()
    def make_message(name: str, age: int) -> str:
        return "{} is {}".format(name, age)

    @task()
    def main() -> str:
        name, age = make_pair()
        return make_message(name, age)

    @task()
    def main2() -> str:
        pair = make_pair()
        return make_message(pair[0], pair[1])

    assert scheduler.run(main()) == "Bob is 20"
    assert scheduler.run(main2()) == "Bob is 20"


def test_nout(scheduler: Scheduler) -> None:
    """
    Tasks with nout should support iteration.
    """

    @task(nout=3)
    def make_list() -> List[int]:
        return [1, 2, 3]

    @task()
    def process(x):
        return x + 1

    @task()
    def main() -> List[int]:
        # make_list() supports iteration because of nout.
        return [process(x) for x in make_list()]

    assert scheduler.run(main()) == [2, 3, 4]


def test_nout_undefined(scheduler: Scheduler) -> None:
    """
    Tasks without nout should not support iteration.
    """

    @task()
    def make_list() -> List[int]:
        return [1, 2, 3]

    with pytest.raises(TypeError):
        for x in make_list():
            pass


def test_nout_bad(scheduler: Scheduler) -> None:
    """
    Task nout must be non-negative int.
    """
    with pytest.raises(TypeError):

        @task(nout="3")
        def make_list() -> List[int]:
            return [1, 2, 3]

    with pytest.raises(TypeError):

        @task(nout=-1)
        def make_list2() -> List[int]:
            return [1, 2, 3]


def test_nout_options(scheduler: Scheduler) -> None:
    """
    Tasks should be able to specify nout using options().
    """

    @task()
    def make_list() -> List[int]:
        return [1, 2, 3]

    @task()
    def process(x):
        return x + 1

    @task()
    def main() -> List[int]:
        # make_list() supports iteration because of nout.
        return [process(x) for x in make_list.options(nout=3)()]

    assert scheduler.run(main()) == [2, 3, 4]


def test_nest_args(scheduler: Scheduler) -> None:
    """
    Tasks should allow nested arguments.
    """

    @task()
    def task1():
        return {
            "output1": 10,
            "output2": 20,
        }

    @task()
    def task2(x):
        return x + 1

    @task()
    def sum_task(xs):
        return sum(xs)

    @task()
    def workflow():
        outputs = task1()
        a = task2(outputs["output1"])
        b = task2(outputs["output2"])
        return sum_task([a, b])

    assert scheduler.run(workflow()) == 32


def test_nested_results(scheduler: Scheduler) -> None:
    """
    Tasks should allow nested results.
    """

    @task()
    def task1():
        return 10

    @task()
    def task2():
        return [task1()]

    result = scheduler.run(task2())
    assert result == [10]


def test_nested_results2(scheduler: Scheduler) -> None:
    @task()
    def to_str(x: int) -> str:
        return str(x)

    def create_linked_list(y: List[int]) -> Optional[Node]:
        if len(y) > 1:
            return Node(value=to_str(y[0]), next_=create_linked_list(y[1:]))
        else:
            return Node(to_str(y[0]), next_=None) if y else None

    assert scheduler.run(create_linked_list([1])) == Node("1", None)
    assert scheduler.run(create_linked_list([1, 2])) == Node("1", Node("2", None))


def test_higher_order(scheduler: Scheduler) -> None:
    """
    Using task as an argument to another task.
    """
    task_calls = []

    @task()
    def get_offset():
        task_calls.append("get_offset")
        return 10

    @task()
    def adder_helper(values):
        return sum(values)

    @task()
    def adder(a, b):
        task_calls.append("adder")
        return adder_helper([a, b, get_offset()])

    @task()
    def executor(my_task, a, b):
        task_calls.append("executor")
        return my_task(a, b)

    @task()
    def workflow():
        task_calls.append("workflow")
        return executor(adder, 2, 3)

    # Ensure that workflow runs correctly.
    assert scheduler.run(workflow()) == 15
    assert task_calls == ["workflow", "executor", "adder", "get_offset"]

    # Ensure cache is reused even when argument is a task.
    task_calls = []
    assert scheduler.run(workflow()) == 15
    assert task_calls == []

    # Changing a child of adder(), changes adder(). Caching should detect this.
    @task()  # type: ignore[no-redef]
    def get_offset():
        task_calls.append("get_offset")
        return 20

    task_calls = []
    assert scheduler.run(workflow()) == 25
    assert task_calls == ["get_offset"]


def test_higher_order2(scheduler: Scheduler) -> None:
    """
    Using task as an output of another task.
    """
    task_calls = []

    @task()
    def helper(values):
        return sum(values)

    @task()
    def helper2(values):
        result = 1
        for value in values:
            result *= value
        return result

    @task()
    def get_offset():
        task_calls.append("get_offset")
        return 10

    @task()
    def adder(a, b):
        task_calls.append("adder")
        return helper([a, b, get_offset()])

    @task()
    def multiplier(a, b):
        task_calls.append("multiplier")
        return helper2([a, b, get_offset()])

    @task()
    def chooser(key):
        task_calls.append("chooser")
        if key == "adder":
            return adder
        elif key == "multiplier":
            return multiplier
        else:
            raise NotImplementedError(key)

    @task()
    def executor(my_task_key, a, b):
        task_calls.append("executor")
        my_task = chooser(my_task_key)
        return my_task(a, b)

    @task()
    def workflow():
        task_calls.append("workflow")
        return helper([executor("adder", 2, 3), executor("multiplier", 2, 3)])

    # Ensure that workflow runs correctly.
    task_calls = []
    assert scheduler.run(workflow()) == 75
    assert sorted(task_calls) == sorted(
        [
            "workflow",
            "executor",
            "chooser",
            "adder",
            "get_offset",
            "executor",
            "chooser",
            "multiplier",
        ]
    )

    # Ensure cache is reused even when argument is a task.
    task_calls = []
    assert scheduler.run(workflow()) == 75
    assert task_calls == []

    # Change definition of task that was cached.
    @task()  # type: ignore[no-redef]
    def adder(a, b):
        task_calls.append("adder")
        return a + b

    # We detect that the cached task is not valid to use, so adder is called again.
    # The rest is still cached.
    task_calls = []
    assert scheduler.run(workflow()) == 65
    assert task_calls == ["chooser", "adder"]

    # Changing a child of multiplier, changes multiplier. Caching should detect this.
    @task()  # type: ignore[no-redef]
    def get_offset():
        task_calls.append("get_offset")
        return 20

    task_calls = []
    assert scheduler.run(workflow()) == 125
    assert task_calls == ["get_offset"]


def test_default_args(scheduler: Scheduler) -> None:
    """
    Changes in default arguments should be considered by caching.
    """
    default = 10

    @task()
    def add(a, b=default):
        return a + b

    assert scheduler.run(add(1, 2)) == 3
    assert scheduler.run(add(1)) == 11

    # Now let's change the default args, without changing the task.
    default = 20

    @task()  # type: ignore
    def add(a, b=default):
        return a + b

    # We should use the new default args.
    assert scheduler.run(add(1)) == 21


def test_novel_kwargs(scheduler: Scheduler) -> None:
    """
    Tasks should be able to accept novel kwargs.
    """

    @task()
    def task1(**kwargs: Any) -> Dict[str, Any]:
        return kwargs

    assert scheduler.run(task1(a=1, b=2)) == {"a": 1, "b": 2}


def test_no_cache_task(scheduler: Scheduler) -> None:
    """
    Tasks should be able to disable caching.
    """
    task_calls = []

    # Try out the legacy option and all three
    @task(cache_scope=CacheScope.NONE)
    def task1():
        task_calls.append("task1")
        return 10

    task_calls = []
    scheduler.run(task1())
    assert task_calls == ["task1"]

    # Running the task again will execute again because we have disabled caching.
    scheduler.run(task1())
    assert task_calls == ["task1", "task1"]


def test_no_cache(scheduler: Scheduler) -> None:
    """
    Scheduler should be able to disable caching.
    """
    task_calls = []

    @task()
    def task1():
        task_calls.append("task1")
        return 10

    scheduler.run(task1())
    assert task_calls == ["task1"]

    # Running the task again will execute again because we have disabled caching.
    scheduler.run(task1(), cache=False)
    assert task_calls == ["task1", "task1"]


def test_cse_scopes(scheduler: Scheduler) -> None:
    """
    Scheduler should trigger CSE against both pending and completed jobs, depending upon the
    requested scope.
    """
    task_calls = []

    @task
    def noop(any: Any) -> None:
        # Use this no-op to ensure all the expressions are different.
        return

    @task(cache_scope=CacheScope.CSE)
    def task1(any):
        task_calls.append("task1")
        return 10

    @task(cache_scope=CacheScope.NONE)
    def task2():
        return task1(noop(3))

    @task(cache_scope=CacheScope.NONE)
    def task3():
        return task1.options(cache_scope=CacheScope.NONE)(noop(4))

    @task(cache_scope=CacheScope.CSE)
    def task_fail(any):
        task_calls.append("task_fail")
        raise RuntimeError("Fail")

    @task(cache_scope=CacheScope.NONE)
    def task_wrapped_fail():
        return task_fail(noop(2))

    # Spy on the call to `collapse`, as a way to distinguish between a cache hit from the
    # backend and from a pending job.
    with patch.object(Job, "collapse", side_effect=Job.collapse, autospec=True) as collapse_mock:
        # The second call will trigger CSE against the first job while it is still pending
        task_calls = []
        scheduler.run([task1(noop(1)), task1(noop(2))])
        assert collapse_mock.call_count == 1
        assert task_calls == ["task1"]

        # Verify there is an entry in the backend, so that our later tests demonstrate that
        # it's not being retrieved.
        task_calls = []
        scheduler.run([task1.options(cache_scope=CacheScope.BACKEND)])
        assert task_calls == []

        # Running the task again will execute once more, because we can't use the one from the
        # backend.
        collapse_mock.reset_mock()
        task_calls = []
        scheduler.run([task1(noop(1)), task1(noop(2))])
        assert collapse_mock.call_count == 1
        assert task_calls == ["task1"]

        # Ensure that CSE from pending jobs honors the cache setting
        collapse_mock.reset_mock()
        task_calls = []
        scheduler.run([task1(noop(1)), task1.options(cache_scope=CacheScope.NONE)(noop(2))])
        assert collapse_mock.call_count == 0
        assert task_calls == ["task1", "task1"]

        # The second call is delayed, so it will come from the backend, not the pending cache.
        collapse_mock.reset_mock()
        task_calls = []
        scheduler.run([task1(noop(1)), task2()])
        assert collapse_mock.call_count == 0
        assert task_calls == ["task1"]

        # Ensure that CSE from the backend also honors the cache setting
        collapse_mock.reset_mock()
        task_calls = []
        scheduler.run([task1(noop(1)), task3()])
        assert collapse_mock.call_count == 0
        assert task_calls == ["task1", "task1"]

        # Combine the two styles
        collapse_mock.reset_mock()
        task_calls = []
        scheduler.run([task1(noop(1)), task1(noop(2)), task2()])
        assert collapse_mock.call_count == 1
        assert task_calls == ["task1"]

        # Check that errors are replayed from pending CSE
        collapse_mock.reset_mock()
        task_calls = []
        with pytest.raises(RuntimeError, match="Fail"):
            scheduler.run([task_fail(noop(1)), task_fail(noop(2))])
        assert collapse_mock.call_count == 1
        assert task_calls == ["task_fail"]

        # Or from remote CSE.
        collapse_mock.reset_mock()
        task_calls = []
        with pytest.raises(RuntimeError, match="Fail"):
            scheduler.run([task_fail(noop(1)), task_wrapped_fail()])
        assert collapse_mock.call_count == 0
        assert task_calls == ["task_fail"]


class InvalidValue(Value):
    """A value that is never valid, hence can't be cached."""

    def is_valid(self) -> bool:
        return False

    def __setstate__(self, state):
        pass

    def __getstate__(self):
        return {}


def test_cse_no_validity(scheduler: Scheduler) -> None:
    """
    Validity of values is not checked upon CSE.
    """
    task_calls = []

    @task
    def noop():
        # Ensure we don't get any expression cycles.
        return None

    @task()
    def task1(any=None):
        task_calls.append("task1")
        return InvalidValue()

    @task(cache_scope=CacheScope.NONE)
    def task2():
        return task1()

    # Seed the cache
    scheduler.run([task1()])
    assert task_calls == ["task1"]

    # We can't get a hit, because it's not valid.
    task_calls = []
    scheduler.run([task1()])
    assert task_calls == ["task1"]

    # But if we get a CSE hit, we don't check validity, either from a pending job
    task_calls = []
    scheduler.run([task1(), task1(any=noop())])
    assert task_calls == ["task1"]

    # or from the backend.
    task_calls = []
    scheduler.run([task1(), task2()])
    assert task_calls == ["task1"]


def test_job_options() -> None:
    """
    The task options for a job should follow the precedence of:
    1. Expression level.
    2. Task level.
    """

    @task(option1="aaa", option2="bbb")
    def task1():
        return 10

    expr = task1.options(option2="ccc")()
    job = Job(task1, expr)
    job.task = task1
    assert job.get_options() == {
        "option1": "aaa",
        "option2": "ccc",
    }


def test_job_status() -> None:
    """
    Job status should update as job progresses.
    """

    @task()
    def task1():
        return 10

    job = Job(task1, task1())
    assert job.status == "PENDING"

    job.eval_args = ((), {})
    assert job.status == "RUNNING"

    job.resolve(10)
    assert job.status == "DONE"

    job = Job(task1, task1())
    job.eval_args = ((), {})
    job.reject(ValueError())
    assert job.status == "FAILED"


def test_log_job_status(scheduler: Scheduler) -> None:
    """
    Scheduler should display a Job status table.
    """

    @task()
    def task1(x):
        return x + 1

    @task()
    def main():
        return [task1(1), task1(2)]

    assert scheduler.run(main()) == [2, 3]

    logs = []

    def log(*messages, **args):
        logs.extend(messages)

    with patch.object(scheduler, "log", wraps=log):
        report = scheduler.get_job_status_report()
        scheduler.log_job_statuses()

    assert report == logs
    assert logs[1:] == [
        "| TASK    PENDING RUNNING  FAILED  CACHED    DONE   TOTAL",
        "| ",
        "| ALL           0       0       0       0       3       3",
        "| main          0       0       0       0       1       1",
        "| task1         0       0       0       0       2       2",
        "",
    ]


def test_dryrun(scheduler: Scheduler) -> None:
    """
    No jobs should be submitted to executors during dryrun.
    """

    @task()
    def task1():
        return 10

    def boom(*args):
        raise AssertionError()

    # No jobs should be submitted to executor.
    with patch("redun.executors.local.LocalExecutor.submit") as submit_mock:
        submit_mock.side_effect = boom

        with pytest.raises(DryRunResult):
            scheduler.run(task1(), dryrun=True)
            assert not submit_mock.called

    # Run workflow.
    assert scheduler.run(task1()) == 10

    with patch("redun.executors.local.LocalExecutor.submit") as submit_mock:
        submit_mock.side_effect = boom

        # Workflow will now be fully cached.
        assert scheduler.run(task1()) == 10
        assert not submit_mock.called


def test_dryrun_explain(scheduler: Scheduler) -> None:
    """
    Dryrun should log explanation for cache miss.
    """

    @task(version="1")
    def main(x):
        return x

    # Run pipeline.
    assert scheduler.run(main(10)) == 10

    # Update arguments.
    with patch.object(scheduler.logger, "log") as log:
        with pytest.raises(DryRunResult):
            scheduler.run(main(11), dryrun=True)

        assert any(
            "Existing task 'main()' is called with new arguments" in call[0][1]
            for call in log.call_args_list
        )

    # Update task.
    @task(version="2")  # type: ignore
    def main(x):
        return x

    with patch.object(scheduler.logger, "log") as log:
        with pytest.raises(DryRunResult):
            scheduler.run(main(10), dryrun=True)

        assert any(
            "New task 'main()' with previous arguments" in call[0][1]
            for call in log.call_args_list
        )

    # Update arguments and task.
    with patch.object(scheduler.logger, "log") as log:
        with pytest.raises(DryRunResult):
            scheduler.run(main(11), dryrun=True)

        assert any(
            "New task 'main()' is called with new arguments" in call[0][1]
            for call in log.call_args_list
        )


def test_log_error_stack(scheduler: Scheduler) -> None:
    """
    Scheduler should log the task stack when raising an exception.
    """

    def func():
        raise ValueError("Boom")

    @task()
    def task1(a, b, c=3):
        return func()

    @task()
    def main(x):
        return task1(x, x + 1)

    scheduler.logger = Mock()

    with pytest.raises(ValueError):
        scheduler.run(main(10))

    logs = "\n".join(call[1][1] for call in scheduler.logger.method_calls)
    traceback_logs = logs.split("\n")[-14:]

    expected_logs = [
        r"\*\*\* Execution failed. Traceback \(most recent task last\):",
        r'  Job .*: File ".*", line \d+, in main',
        r"    def main\(x\):",
        r"    x = 10",
        r'  Job .*: File ".*", line \d+, in task1',
        r"    def task1\(a, b, c=3\):",
        r"    a = 10",
        r"    b = 11",
        r"    c = 3",
        r'  File ".*", line \d+, in task1',
        r"    return func\(\)",
        r'  File ".*", line \d+, in func',
        r'    raise ValueError\("Boom"\)',
        r"ValueError: Boom",
    ]
    assert_match_lines(expected_logs, traceback_logs)


def test_traceback_trim_frames() -> None:
    """
    Traceback should be able to trim frames related to redun scheduler.
    """
    basedir = os.path.dirname(os.path.dirname(__file__))

    @task()
    def task1():
        return 10

    frames = Traceback.trim_frames(
        [
            FrameSummary(basedir + "/__init__.py", 5, "func0"),
            FrameSummary(basedir + "/executors/local.py", 5, "func0"),
            Frame("myfile.py", 10, "func", {}, job=Job(task1, task1())),
            Frame("myfile2.py", 20, "func2", {}, job=Job(task1, task1())),
        ]
    )
    assert [frame.filename for frame in frames] == ["myfile.py", "myfile2.py"]

    frames = Traceback.trim_frames(
        [
            FrameSummary(basedir + "/cli.py", 5, "func0"),
            Frame("myfile.py", 10, "func", {}, job=Job(task1, task1())),
            Frame("myfile2.py", 20, "func2", {}, job=Job(task1, task1())),
        ]
    )
    assert [frame.filename for frame in frames] == ["myfile.py", "myfile2.py"]


def test_traceback_serialize() -> None:
    """
    Traceback should be able to trim frames related to redun scheduler.
    """
    basedir = os.path.dirname(os.path.dirname(__file__))

    @task()
    def task1():
        return 10

    traceback = Traceback(
        error=ValueError("boom"),
        frames=[
            FrameSummary(basedir + "/__init__.py", 5, "func0"),
            FrameSummary(basedir + "/executors/local.py", 5, "func0"),
            Frame("myfile.py", 10, "func", {}, job=Job(task1, task1())),
            Frame("myfile2.py", 20, "func2", {}, job=Job(task1, task1())),
        ],
        logs=[
            "line1",
            "line2",
            "line3",
        ],
    )

    # Traceback and Frame should have a type_name.
    assert Frame.type_name == "redun.Frame"
    assert Traceback.type_name == "redun.Traceback"

    # Traceback and Frame should serialize and deserialize.
    registry = get_type_registry()
    data = registry.serialize(traceback)
    traceback2 = registry.deserialize("redun.Traceback", data)
    data2 = registry.serialize(traceback2)

    assert data == data2


def test_check_valid(scheduler: Scheduler, session: Session) -> None:
    """
    check_valid=shallow should skip evaluating subtrees of the call graph.
    """
    calls = []

    @task()
    def task1():
        calls.append("task1")
        return 10

    @task()
    def main():
        return task1()

    with patch.object(Scheduler, "_exec_job", wraps=scheduler._exec_job) as exec_job_mock:
        assert scheduler.run(main.options(check_valid="full")()) == 10
        assert calls == ["task1"]

        task_execs = [call[0][0].task.name for call in exec_job_mock.call_args_list]
        assert task_execs == ["main", "task1"]

    with patch.object(Scheduler, "_exec_job", wraps=scheduler._exec_job) as exec_job_mock:
        assert scheduler.run(main.options(check_valid="full")()) == 10
        # task1 is not executed again, due to caching.
        assert calls == ["task1"]

        # But we do recurse to task1 to assess its cache value.
        task_execs = [call[0][0].task.name for call in exec_job_mock.call_args_list]
        assert task_execs == ["main", "task1"]

    with patch.object(Scheduler, "_exec_job", wraps=scheduler._exec_job) as exec_job_mock:
        # Now use check_valid="shallow".
        assert scheduler.run(main.options(check_valid="shallow")()) == 10
        assert calls == ["task1"]

        # We short circuit the cache checking at the top-level main task.
        task_execs = [call[0][0].task.name for call in exec_job_mock.call_args_list]
        assert task_execs == ["main"]

    # We should have recorded the same CallNode for each execution.
    from redun.backends.db import Execution

    executions = session.query(Execution).all()
    assert len({execution.job.call_hash for execution in executions}) == 1


def test_scheduler_task(scheduler: Scheduler) -> None:
    """
    scheduler_task decorator should allow custom evaluation.
    """

    @scheduler_task("task1", "redun")
    def task1(
        scheduler: Scheduler, parent_job: Job, sexpr: SchedulerExpression, x: int
    ) -> Promise:
        return scheduler.evaluate(x, parent_job=parent_job).then(lambda x2: x2 + 1)

    expr = task1(1)
    assert isinstance(task1, SchedulerTask)
    assert isinstance(expr, SchedulerExpression)
    assert expr.task_name == "redun.task1"
    assert expr.args == (1,)

    assert scheduler.run(task1(1)) == 2
    assert scheduler.run(task1(task1(1))) == 3

    # SchedulerTasks should be first-class values.
    assert isinstance(task1, Value)

    # SchedulerTasks should support partial application.
    expr2 = task1.partial()
    assert isinstance(expr2, PartialTask)
    assert isinstance(expr2(1), SchedulerExpression)
    assert scheduler.run(task1.partial()(1)) == 2


def test_cond(scheduler: Scheduler) -> None:
    """
    cond() should act like lasy if-statement.
    """

    @task()
    def id(x):
        return x

    @task()
    def boom():
        raise ValueError()

    assert scheduler.run(cond(id(True), id(1), id(2))) == 1
    assert scheduler.run(cond(id(False), id(1), id(2))) == 2

    # boom should not execute at all.
    assert scheduler.run(cond(id(False), boom(), id(2))) == 2

    # if, elif, else
    assert scheduler.run(cond(id(False), id(1), id(True), id(2), id(True), id(3))) == 2


def test_reduce(scheduler: Scheduler) -> None:
    @scheduler_task(namespace="redun.tests.test_scheduler")
    def reduce_(
        scheduler: Scheduler,
        parent_job: Job,
        sexpr: SchedulerExpression,
        a_task: Task,
        init: Any,
        values: Sequence[Any],
    ) -> Promise:
        def _reduce(args: List):
            a_task, init, values = args
            reduce_promise: Promise = Promise()
            queue = values if values else [init]
            remaining = len(values) - 1

            def fail(error):
                reduce_promise.do_reject(error)

            def then(value):
                nonlocal remaining
                remaining -= 1

                if remaining == 0:
                    reduce_promise.do_resolve(value)
                    return value

                queue.append(value)

                while len(queue) >= 2:
                    value1 = queue.pop()
                    value2 = queue.pop()
                    scheduler.evaluate(a_task(value1, value2), parent_job=parent_job).then(
                        then, fail
                    )

            if len(queue) == 1:
                # Base case.
                reduce_promise.do_resolve(queue[0])
            else:
                value1 = queue.pop()
                value2 = queue.pop()
                scheduler.evaluate(a_task(value1, value2), parent_job=parent_job).then(then, fail)

            return reduce_promise

        return scheduler.evaluate([a_task, init, values]).then(_reduce)

    @task()
    def add(a, b):
        return a + b

    assert scheduler.run(reduce_(add, 0, [])) == 0
    assert scheduler.run(reduce_(add, 0, [1])) == 1
    assert scheduler.run(reduce_(add, 0, [1, 2])) == 3
    assert scheduler.run(reduce_(add, 0, [1, 2, 3])) == 6
    assert scheduler.run(reduce_(add, 0, list(range(20)))) == 190


def test_catch(scheduler: Scheduler) -> None:
    """
    Catch expression should handle exceptions.
    """

    @task()
    def faulty():
        return 1 / 0

    @task()
    def recover(error):
        return 1.0

    @task()
    def task1():
        return catch(faulty(), ZeroDivisionError, recover)

    @task()
    def task2():
        return faulty()

    @task()
    def task3():
        # Catch should also allow multiple error classes.
        return catch(faulty(), (ZeroDivisionError, ValueError), recover)

    # Single task should be able to raise an exception.
    with pytest.raises(ZeroDivisionError):
        scheduler.run(faulty())

    # A nested task should be able to raise an exception.
    with pytest.raises(ZeroDivisionError):
        scheduler.run(task2())

    # A catch expression should catch the exception.
    assert scheduler.run(task1()) == 1.0
    assert scheduler.run(task3()) == 1.0


def test_multi_catch(scheduler: Scheduler) -> None:
    """
    Catch expression should handle multiple exception classes.
    """

    @task()
    def faulty():
        return 1 / 0

    @task()
    def faulty2():
        return {"a": 1}["b"]

    @task()
    def recover1(error):
        return "hello"

    @task()
    def recover2(error):
        return 1.0

    @task()
    def task1():
        return catch(
            faulty(),
            KeyError,
            recover1,
            ZeroDivisionError,
            recover2,
        )

    assert scheduler.run(task1()) == 1.0

    @task()
    def task2():
        return catch(
            faulty2(),
            KeyError,
            recover1,
            ZeroDivisionError,
            recover2,
        )

    assert scheduler.run(task2()) == "hello"


def test_catch_reraise(scheduler: Scheduler) -> None:
    """
    Catch expression can reraise a different exception.
    """

    @task()
    def faulty():
        return 1 / 0

    @task()
    def recover(error):
        raise ValueError("Reraised exception {}.".format(error))

    @task()
    def task1():
        return catch(faulty(), ZeroDivisionError, recover)

    with pytest.raises(ValueError):
        scheduler.run(task1())


def test_catch_cache(scheduler: Scheduler) -> None:
    """
    Catch expression should cache if successful.
    """

    @task()
    def faulty():
        calls.append("faulty")
        return 1 / 0

    @task()
    def recover(error):
        return 1.0

    @task()
    def reraise(error):
        raise error

    calls: List[str] = []
    with pytest.raises(ZeroDivisionError):
        scheduler.run(faulty())

    with pytest.raises(ZeroDivisionError):
        scheduler.run(faulty())

    # Failed tasks should not cache.
    assert calls == ["faulty", "faulty"]

    # Failed catch should not cache.
    calls = []
    assert scheduler.run(catch(faulty(), ZeroDivisionError, recover)) == 1.0
    assert calls == ["faulty"]

    # Previously recovered catch should cache.
    assert scheduler.run(catch(faulty(), ZeroDivisionError, recover)) == 1.0
    assert calls == ["faulty"]

    # Caching should not be used if turned off at the scheduler level
    assert scheduler.run(catch(faulty(), ZeroDivisionError, recover), cache=False) == 1.0
    assert calls == ["faulty", "faulty"]

    # Caching should not be used if turned off at the task level.
    assert (
        scheduler.run(
            catch(faulty.options(cache_scope=CacheScope.NONE)(), ZeroDivisionError, recover),
            cache=False,
        )
        == 1.0
    )
    assert calls == ["faulty", "faulty", "faulty"]

    # Reraised catch should not cache.
    calls = []
    with pytest.raises(ZeroDivisionError):
        scheduler.run(catch(faulty(), ZeroDivisionError, reraise))
    assert calls == ["faulty"]

    with pytest.raises(ZeroDivisionError):
        scheduler.run(catch(faulty(), ZeroDivisionError, reraise))
    assert calls == ["faulty", "faulty"]


def test_catch_task_react(scheduler: Scheduler) -> None:
    """
    Catch should react to new task hashes.
    """

    @task()
    def safe():
        calls.append("safe")
        return 1 / 1

    @task()
    def faulty():
        calls.append("faulty")
        return 1 / 0

    @task()
    def recover(error):
        calls.append("recover")
        return 1.0

    # Running safe twice should cache.
    calls: List[str] = []
    assert scheduler.run(catch(safe(), ZeroDivisionError, recover)) == 1
    assert calls == ["safe"]
    assert scheduler.run(catch(safe(), ZeroDivisionError, recover)) == 1
    assert calls == ["safe"]

    # Updating safe task should force re-execution.
    @task()  # type: ignore[no-redef]
    def safe():
        calls.append("safe2")
        return 2 / 1

    assert scheduler.run(catch(safe(), ZeroDivisionError, recover)) == 2
    assert calls == ["safe", "safe2"]

    # Running recover twice should cache.
    calls = []
    assert scheduler.run(catch(faulty(), ZeroDivisionError, recover)) == 1.0
    assert calls == ["faulty", "recover"]
    assert scheduler.run(catch(faulty(), ZeroDivisionError, recover)) == 1.0
    assert calls == ["faulty", "recover"]

    # Updating recover task should force re-execution.
    @task()  # type: ignore[no-redef]
    def recover(error):
        calls.append("recover2")
        return 2.0

    assert scheduler.run(catch(faulty(), ZeroDivisionError, recover)) == 2.0
    assert calls == ["faulty", "recover", "faulty", "recover2"]


def test_catch_deep_task_react(scheduler: Scheduler) -> None:
    """
    Catch should react to new task hashes deep in the workflow.
    """

    @task()
    def deep(denom):
        calls.append("deep")
        return 1 / denom

    @task()
    def safe():
        calls.append("safe")
        return deep(1)

    @task()
    def faulty():
        calls.append("faulty")
        return 1 / 0

    @task()
    def recover(error):
        calls.append("recover")
        return 1.0

    # Running safe twice should cache.
    calls: List[str] = []
    assert scheduler.run(catch(safe(), ZeroDivisionError, recover)) == 1
    assert calls == ["safe", "deep"]
    assert scheduler.run(catch(safe(), ZeroDivisionError, recover)) == 1
    assert calls == ["safe", "deep"]

    # Updating deep task should force re-execution.
    @task()  # type: ignore[no-redef]
    def deep(denom):
        calls.append("deep2")
        return 2 / denom

    assert scheduler.run(catch(safe(), ZeroDivisionError, recover)) == 2
    assert calls == ["safe", "deep", "deep2"]


def test_catch_deep_recover_react(scheduler: Scheduler) -> None:
    """
    Catch should react to new task hashes deep in the recover workflow.
    """

    @task()
    def deep():
        calls.append("deep")
        return 1

    @task()
    def safe():
        calls.append("safe")
        return 1 / 1

    @task()
    def faulty():
        calls.append("faulty")
        return 1 / 0

    @task()
    def recover(error):
        calls.append("recover")
        return deep()

    # Running recover twice should cache.
    calls: List[str] = []
    assert scheduler.run(catch(faulty(), ZeroDivisionError, recover)) == 1
    assert calls == ["faulty", "recover", "deep"]
    assert scheduler.run(catch(faulty(), ZeroDivisionError, recover)) == 1
    assert calls == ["faulty", "recover", "deep"]

    # Updating deep task should force re-execution.
    @task()  # type: ignore[no-redef]
    def deep():
        calls.append("deep2")
        return 2

    assert scheduler.run(catch(faulty(), ZeroDivisionError, recover)) == 2
    assert calls == ["faulty", "recover", "deep", "deep2"]


def test_catch_all(scheduler: Scheduler) -> None:
    """
    Catch expression should handle exceptions.
    """

    @task
    def throw(error: Exception) -> None:
        raise error

    @task
    def good(x=10):
        return x

    @task
    def bad(message="boom"):
        raise ValueError(message)

    @task
    def recover(results):
        return map_nested_value(
            lambda result: (0 if isinstance(result, Exception) else result), results
        )

    # All subtasks succeed.
    assert scheduler.run(catch_all([good(1), good(2), good(3)], ValueError, recover)) == [1, 2, 3]

    # Failed tasks should be caught.
    assert scheduler.run(catch_all([good(4), bad(), good(5)], ValueError, recover)) == [4, 0, 5]
    assert scheduler.run(catch_all([good(6), bad(), bad(), good(7)], ValueError, recover)) == [
        6,
        0,
        0,
        7,
    ]

    # We should be able to catch multiple error classes.
    assert scheduler.run(
        catch_all(
            [throw(KeyError()), throw(ValueError()), good()], (KeyError, ValueError), recover
        )
    ) == [0, 0, 10]

    # Without a recovery, errors should reraise.
    with pytest.raises(ValueError):
        scheduler.run(catch_all([good(), bad(), good()]))

    # Different error_class should pass through.
    with pytest.raises(ValueError):
        scheduler.run(catch_all([good(), bad(), good()], ZeroDivisionError, recover))

    # Different error_class should pass through, even if others are caught.
    with pytest.raises(ValueError):
        scheduler.run(
            catch_all([throw(KeyError()), throw(ValueError()), good()], KeyError, recover)
        )

    # Any kind of nested value should work.
    assert scheduler.run(catch_all([good(8), [bad(), {"a": good(9)}]], ValueError, recover)) == [
        8,
        [0, {"a": 9}],
    ]


def test_config_args(scheduler: Scheduler) -> None:
    """
    Config_args should not contribute to the eval_hash.
    """

    @task(config_args=["x", "z"])
    def task1(x: int, y: int, z: str = "hello") -> int:
        return x

    eval_hash1, args_hash1 = scheduler.get_eval_hash(task1, (10, 11), {})
    # Changing 'x' and 'z' does not change args_hash.
    eval_hash2, args_hash2 = scheduler.get_eval_hash(task1, (11, 11), {"z": "bye"})
    assert eval_hash1 == eval_hash2
    assert args_hash1 == args_hash2


def test_variadic_args(scheduler: Scheduler) -> None:
    """
    Variadic arguments should contribute to the eval_hash.
    """

    @task
    def task1(x: int, y: int, *rest: int) -> int:
        return x

    eval_hash1, args_hash1 = scheduler.get_eval_hash(task1, (10, 11, 12), {})
    eval_hash2, args_hash2 = scheduler.get_eval_hash(task1, (10, 11, 12, 13, 14), {})
    assert eval_hash1 != eval_hash2
    assert args_hash1 != args_hash2


def test_default_root(scheduler: Scheduler, session: Session) -> None:
    """
    We should always have one root job for an Execution.
    """

    @task()
    def task1(x: int) -> int:
        return x

    assert scheduler.run(task1(1) + task1(2)) == 3

    exec1 = session.query(Execution).one()

    # We should have the default root task.
    assert exec1.job.task.fullname == "redun.root_task"

    # The default root should have two child jobs.
    assert len(exec1.job.child_jobs) == 2


def test_no_default_root(scheduler: Scheduler, session: Session) -> None:
    """
    We should not use a default root task if top-level expression is a TaskExpression.
    """

    @task()
    def task1(x: int) -> int:
        return x

    assert scheduler.run(task1(1)) == 1

    exec1 = session.query(Execution).one()

    # We should have the default root task.
    assert exec1.job.task.name == "task1"


def test_default_args_expression(scheduler: Scheduler) -> None:
    """
    Task default arguments should support expressions.
    """

    @task
    def add(a: int, b: int) -> int:
        return a + b

    @task
    def main(x: int = add(1, 2)) -> int:
        return x

    assert scheduler.run(main()) == 3
    assert scheduler.run(main(4)) == 4

    @task
    def main2(x: List[int] = [add(1, 2)]) -> List[int]:
        # Default argument might contain an expression within a nested value (e.g. list).
        return x

    assert scheduler.run(main2()) == [3]


def test_job_clear(scheduler: Scheduler) -> None:
    """
    Job clean up should occur correctly even when child jobs fail.

    https://insitro.atlassian.net/browse/DE-4645
    """

    @task
    def child(x: int) -> int:
        if x == 2:
            raise ValueError("raised by child")
        return x

    @task
    def child2(x: int) -> int:
        return x

    @task
    def parent(n: int) -> List[int]:
        return [child2(child(i)) for i in range(n)]

    @task
    def main(n: int) -> List[int]:
        return parent(n)

    with pytest.raises(ValueError, match="raised by child"):
        scheduler.run(main(10))


def test_common_expression() -> None:
    """
    Scheduler should use Common Expression Elimination.
    """

    @task
    def id(x: int) -> int:
        return x

    @task
    def add(a: int, b: int) -> int:
        return a + b

    @task
    def boom(x: int) -> int:
        time.sleep(2)  # Ensure there's enough time for the pending job hit.
        raise ValueError("BOOM!")

    @task()
    def noop(any):
        """No-op used to ensure expressions are different."""
        return 1

    @task
    def wrapped_boom():
        return boom(noop(2))

    # There should be two jobs, although one will be cached
    scheduler = Scheduler()
    assert isinstance(scheduler.backend, RedunBackendDb)
    assert scheduler.backend.session
    assert scheduler.run([add(1, 2), add(1, 2)]) == [3, 3]
    assert (
        scheduler.backend.session.query(JobDb)
        .filter(JobDb.task_hash == add.hash, JobDb.cached == True)  # noqa: E712
        .count()
        == 0
    )
    assert (
        scheduler.backend.session.query(JobDb)
        .filter(JobDb.task_hash == add.hash, JobDb.cached == False)  # noqa: E712
        .count()
        == 1
    )

    # We should do CSE after argument expressions are evaluated, such as `id(2)`.
    scheduler = Scheduler()
    assert isinstance(scheduler.backend, RedunBackendDb)
    assert scheduler.backend.session

    # Cache id(2) so that we can force `add(1, id(2))` and `add(1, 2)` at the same time
    # and exercise the CSE logic on two different expressions.
    assert scheduler.run(id(2)) == 2
    assert scheduler.run([add(1, id(2)), add(1, 2)]) == [3, 3]
    assert (
        scheduler.backend.session.query(JobDb)
        .filter(JobDb.task_hash == add.hash, JobDb.cached == True)  # noqa: E712
        .count()
        == 1
    )
    assert (
        scheduler.backend.session.query(JobDb)
        .filter(JobDb.task_hash == add.hash, JobDb.cached == False)  # noqa: E712
        .count()
        == 1
    )

    # Expressions that have been collapsed by CSE should also propagate errors.
    scheduler = Scheduler()
    assert isinstance(scheduler.backend, RedunBackendDb)
    assert scheduler.backend.session
    with pytest.raises(ValueError, match="BOOM"):
        # This combination produces a pending CSE hit
        scheduler.run([boom(noop(1)), boom(noop(2))])
    assert (
        scheduler.backend.session.query(JobDb)
        .filter(JobDb.task_hash == boom.hash, JobDb.cached == True)  # noqa: E712
        .count()
        == 1
    )
    assert (
        scheduler.backend.session.query(JobDb)
        .filter(JobDb.task_hash == boom.hash, JobDb.cached == False)  # noqa: E712
        .count()
        == 1
    )


def test_expressions_evaluate_once() -> None:
    """
    Scheduler should not evaluate expression objects more than once.
    """
    calls = []

    @scheduler_task(namespace="redun.tests.test_scheduler")
    def echo(scheduler, parent_job, expr, x) -> Promise:
        calls.append("id")
        promise: Promise[int] = Promise()
        promise.do_resolve(x)
        return promise

    # We should call echo only once, even though scheduler tasks aren't cached.
    scheduler = Scheduler()
    x = echo(1)
    y = x + x
    assert scheduler.run(y) == 2
    assert calls == ["id"]

    # Expression cycle checking will find these as being the same
    calls = []
    y = echo(1) + echo(1)
    assert scheduler.run(y) == 2
    assert calls == ["id"]


def test_expressions_evaluate_once_cache(scheduler: Scheduler, session: Session) -> None:
    """
    Detecting repeated expression objects should work with cached expressions.
    """

    task1_value = 0

    @task(namespace="redun.tests.test_scheduler", cache_scope="NONE")
    def task1():
        nonlocal task1_value
        task1_value += 10
        return task1_value

    @task
    def add(x, y):
        return x + y

    calls = 0

    @task
    def main():
        nonlocal calls
        calls += 1

        x = task1()
        # Reuse the expression object.
        return {
            "a": add(x, 1),
            "b": add(x, 2),
            "c": x,
            "d": task1(),  # This is not a reuse, and should invoke another call.
        }

    # a-c all use the first call, and d gets the second one.
    result = scheduler.run(main())
    assert result["a"] == 11
    assert result["b"] == 12
    assert result["c"] == 10
    assert result["d"] == 10
    assert calls == 1
    assert task1_value == 10

    # If we run it again, `main` will get cached, returning an expression. Within that expression,
    # "a"-"c" share the expression for the nested call to `task1`, hence those get deduplicated.
    # However, "d" is a new expression and is not deduplicated.
    calls = 0
    result2 = scheduler.run(main())
    assert calls == 0  # Check that main was cached
    assert result2["a"] == 21
    assert result2["b"] == 22
    assert result2["c"] == 20
    assert result2["d"] == 20
    assert task1_value == 20


def test_extend_run(scheduler: Scheduler, session: Session) -> None:
    """
    Scheduler should support extending an Execution.
    """

    @task
    def main(x):
        return x + 1

    @task
    def fail():
        raise ValueError("BOOM")

    assert scheduler.run(main(10)) == 11
    job = session.query(JobDb).one()

    result = scheduler.extend_run(main(11), parent_job_id=job.id)
    assert result["result"] == 12

    # Ensure other execution info is present.
    assert result["job_id"]
    assert result["call_hash"]

    # The new job should be a child of the parent job.
    child_job = session.get(JobDb, result["job_id"])
    assert child_job.parent_job.id == job.id
    assert child_job.id in {j.id for j in job.child_jobs}

    # extend_run should also return errors.
    result = scheduler.extend_run(fail(), parent_job_id=job.id)
    assert isinstance(result["error"], ValueError)
    assert str(result["error"]) == "BOOM"
    assert result["job_id"]
    assert result["call_hash"]

    # Force session to refetch new data.
    job_id = job.id
    session.expunge_all()

    # The new job should be a child of the parent job.
    child_job = session.get(JobDb, result["job_id"])
    job = session.get(JobDb, job_id)
    assert child_job.parent_job.id == job.id
    assert child_job.id in {j.id for j in job.child_jobs}


def test_extend_run_dryrun(scheduler: Scheduler, session: Session) -> None:
    """
    Scheduler should support extending an Execution with dryrun.
    """

    @task
    def main(x):
        return x + 1

    @task
    def fail():
        raise ValueError("BOOM")

    assert scheduler.run(main(10)) == 11
    job = session.query(JobDb).one()

    # We should have an incomplete run due to dryrun=True.
    result = scheduler.extend_run(main(11), parent_job_id=job.id, dryrun=True)
    assert result["dryrun"]

    # If we finish the run and then do dryrun, we should get a complete run due to caching.
    result = scheduler.extend_run(main(11), parent_job_id=job.id)
    result = scheduler.extend_run(main(11), parent_job_id=job.id, dryrun=True)
    assert result["result"] == 12



from redun.executors.alias import AliasExecutor
from redun.executors.aws_batch import AWSBatchExecutor
from redun.executors.aws_glue import AWSGlueExecutor
from redun.executors.docker import DockerExecutor

try:
    from redun.executors.k8s import K8SExecutor
except ModuleNotFoundError:
    # Skip k8s executor if kubernetes is not installed.
    pass
from redun.executors.local import LocalExecutor
from redun.file import Dir, File, ShardedS3Dataset
from redun.handle import Handle
from redun.namespace import get_current_namespace, namespace
from redun.scheduler import (
    Scheduler,
    apply_tags,
    catch,
    cond,
    get_current_scheduler,
    merge_handles,
    set_current_scheduler,
    throw,
)
from redun.scripting import script
from redun.task import PartialTask, Task, get_task_registry, task

__version__ = "0.13.0"



import pickle

import pytest
import sqlalchemy as sa
from sqlalchemy.orm import Session

from redun import Scheduler, task
from redun.backends.db import Execution, Job
from redun.scheduler import ErrorValue
from redun.utils import pickle_dumps
from redun.value import get_type_registry


def test_serializing_lambdas(scheduler):
    """
    redun does not support serializing lambdas.
    """
    # Should raise error.
    # with pytest.raises(Exception):
    #    result = scheduler.run(lambda x: 10)

    # Try to pass lambda as an argument.
    @task()
    def task1(x):
        return x

    with pytest.raises(Exception):
        scheduler.run(task1(lambda: 10))

    # Try to return lambda as a result.
    @task()
    def task2(x):
        return lambda: 10

    with pytest.raises(Exception):
        scheduler.run(task2(10))


def func1(x):
    return x


def is_naked_function(value):
    """
    Return true if value is a function but not a Task.
    """
    return type(value).__name__ == "function"


def test_detecting_func():
    """
    Ensure we can distinguish naked functions from tasks.
    """

    @task()
    def task1(x):
        return x

    assert callable(func1)
    assert is_naked_function(func1)

    assert callable(task1)
    assert not is_naked_function(task1)


def test_serializing_functions(scheduler):
    """
    redun does not support serializing functions.

    Instead users should pass and return Tasks.
    """
    # Functions defined globally can be serialized by pickle.
    # The function bytecode not serialized, just a name to the func.
    data = pickle_dumps(func1)
    assert data == b"\x80\x03credun.tests.test_errors\nfunc1\nq\x00."

    # If we change the function name to 'bad_func', pickle will raise an
    # error when deserializing.
    data2 = b"\x80\x03credun.tests.test_errors\nbad_func\nq\x00."
    with pytest.raises(AttributeError):
        pickle.loads(data2)

    # Try to pass a function as an argument.
    @task()
    def task1(x):
        return x

    # with pytest.raises(Exception):
    #    scheduler.run(task1(func1))


def test_exceptions(scheduler: Scheduler) -> None:
    """
    Ensure exceptions in tasks are handled correctly.
    """

    @task()
    def faulty():
        raise ValueError("BOOM")

    @task()
    def test1(x):
        return x

    @task()
    def test2():
        return faulty()

    # Task that raises an exception should be raised from run().
    with pytest.raises(ValueError):
        scheduler.run(faulty())

    # Error raised in Task as argument should be raised from run().
    with pytest.raises(ValueError):
        scheduler.run(test1(faulty()))

    # Error raised in Task as result should be raised from run().
    with pytest.raises(ValueError):
        scheduler.run(test2())


def test_exception_call_graph(scheduler: Scheduler, session: Session) -> None:
    """
    Exceptions should be recorded in call graph.
    """

    @task(version="1")
    def faulty():
        raise ValueError("BOOM")

    @task()
    def inc(x):
        return x + 1

    @task()
    def task1():
        return [inc(10), faulty()]

    # Workflow should raise exception.
    with pytest.raises(ValueError):
        scheduler.run(task1())

    # CallGraph should have recorded exception as ValueError.
    execution = (
        session.query(Execution)
        .join(Job, Job.id == Execution.job_id)
        .order_by(sa.desc(Job.start_time))
    ).first()
    assert execution.job.call_node
    assert isinstance(execution.job.call_node.value.value_parsed.error, ValueError)

    # Both successful and failed jobs should be recorded.
    assert execution.job.child_jobs[0].call_node.value.value_parsed == 11
    assert execution.job.child_jobs[1].call_node.value.value_parsed.error.args == ("BOOM",)

    # Exceptions should not be cached. Let's fix the bug without bump version.

    @task(version="1")  # type: ignore[no-redef]
    def faulty():
        return 20

    assert scheduler.run(task1.options(check_valid="shallow")()) == [11, 20]


def test_error_value() -> None:
    """
    ErrorValue should implement the Value interface.
    """

    @task(version="1")
    def faulty():
        raise ValueError("BOOM")

    @task()
    def test1(x):
        return x

    @task()
    def test2():
        return faulty()

    error = ErrorValue(ValueError("BOOM"))
    assert error.type_name == "redun.ErrorValue"
    assert error.get_hash() == "4e7d45fe277f8638271781336996591c1cf6b7bb"

    registry = get_type_registry()
    bytes = registry.serialize(error)
    error2 = registry.deserialize("redun.ErrorValue", bytes)

    assert isinstance(error2.error, ValueError)
    assert error2.error.args == ("BOOM",)


def test_error_value_fallback(scheduler: Scheduler, session: Session) -> None:
    """
    redun should fallback to Exception for errors that cannot be serialized.
    """

    @task()
    def main():
        # Lambdas are not pickleable.
        raise ValueError(lambda: "unpicklable")

    # Workflow should raise exception.
    with pytest.raises(ValueError):
        scheduler.run(main())

    # CallGraph should have recorded exception as ValueError.
    execution = session.query(Execution).one()
    error = execution.job.call_node.value.value_parsed.error
    assert type(error) == Exception
    assert "ValueError" in str(error)



import json
import logging
import os
import re
import threading
import time
import uuid
from collections import OrderedDict, defaultdict
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, TypeVar, Union, cast

import kubernetes
from kubernetes.client import CoreV1Api, V1Job, V1Pod
from kubernetes.client.exceptions import ApiException

from redun.executors import aws_utils, k8s_utils
from redun.executors.base import Executor, register_executor
from redun.executors.code_packaging import package_code, parse_code_package_config
from redun.executors.command import get_oneshot_command, get_script_task_command
from redun.executors.scratch import (
    SCRATCH_ERROR,
    SCRATCH_HASHES,
    SCRATCH_INPUT,
    SCRATCH_OUTPUT,
    get_array_scratch_file,
    get_job_scratch_dir,
    get_job_scratch_file,
    parse_job_error,
    parse_job_result,
)
from redun.file import File
from redun.job_array import JobArrayer
from redun.logging import logger
from redun.scheduler import Job, Scheduler
from redun.scripting import get_task_command
from redun.task import CacheScope, Task
from redun.utils import pickle_dump

T = TypeVar("T")
SUCCEEDED = "SUCCEEDED"
FAILED = "FAILED"
ARRAY_JOB_SUFFIX = "array"


def k8s_submit(
    k8s_client: k8s_utils.K8SClient,
    command: List[str],
    image: str,
    namespace: str,
    job_name: str = "k8s-job",
    array_size: int = 0,
    memory: int = 4,
    vcpus: int = 1,
    gpus: int = 0,
    timeout: Optional[int] = None,
    k8s_labels: Optional[Dict[str, str]] = None,
    retries: int = 1,
    service_account_name: str = "default",
    annotations: Optional[Dict[str, str]] = None,
    secret_name: Optional[str] = None,
) -> kubernetes.client.V1Job:
    """Prepares and submits a k8s job to the API server"""
    requests = {
        "memory": f"{memory}G",
        "cpu": vcpus,
        "nvidia.com/gpu": gpus,
    }
    limits = requests
    resources = k8s_utils.create_resources(requests, limits)

    k8s_job = k8s_utils.create_job_object(
        name=job_name,
        image=image,
        command=command,
        resources=resources,
        timeout=timeout,
        labels=k8s_labels,
        service_account_name=service_account_name,
        annotations=annotations,
        secret_name=secret_name,
    )

    if array_size > 1:
        k8s_job.spec.completions = array_size
        k8s_job.spec.parallelism = array_size
        k8s_job.spec.completion_mode = "Indexed"
        k8s_job.spec.backoff_limit = retries * array_size
    else:
        k8s_job.spec.backoff_limit = retries
    k8s_job.spec.restart_policy = "OnFailure"

    return k8s_utils.create_job(k8s_client, k8s_job, namespace=namespace)


def get_k8s_job_name(prefix: str, job_hash: str, array: bool = False) -> str:
    """
    Return a K8S Job name by either job or job hash.
    """
    return "{}-{}{}".format(prefix, job_hash, f"-{ARRAY_JOB_SUFFIX}" if array else "")


def is_array_job_name(job_name: str) -> bool:
    "Returns true if the job name looks like an array job"
    return job_name.endswith(f"-{ARRAY_JOB_SUFFIX}")


def get_hash_from_job_name(job_name: str) -> Optional[str]:
    """
    Returns the job/task eval_hash that corresponds with a particular job name
    on K8S.
    """
    # Remove array job suffix, if present.
    array_suffix = "-" + ARRAY_JOB_SUFFIX
    if job_name.endswith(array_suffix):
        job_name = job_name[: -len(array_suffix)]

    # It's possible we found jobs that are unrelated to the this work based off
    # the job_name_prefix matching when fetching in get_jobs. These jobs will
    # not have hashes so we can ignore them. For a concrete example of this,
    # see:
    #
    #   https://insitro.atlassian.net/browse/DE-2632
    #
    # where a headnode job is running but has no hash so we don't want to
    # interact with that job here. If we don't find a match, consider this a
    # case of the above where we matched unrelated jobs and return None to let
    # callers know this is the case.
    match = re.match(".*-(?P<hash>[^-]+)", job_name)
    if match:
        return match["hash"]

    return None


def get_k8s_job_options(job_options: dict) -> dict:
    """
    Returns K8S-specific job options from general job options.
    """
    keys = [
        "memory",
        "vcpus",
        "gpus",
        "k8s_labels",
        "annotations",
        "service_account_name",
        "retries",
        "timeout",
    ]
    return {key: job_options[key] for key in keys if key in job_options}


def submit_task(
    k8s_client: k8s_utils.K8SClient,
    image: str,
    namespace: str,
    scratch_prefix: str,
    job: Job,
    a_task: Task,
    args: Tuple = (),
    kwargs: Dict[str, Any] = {},
    job_options: dict = {},
    array_uuid: Optional[str] = None,
    array_size: int = 0,
    code_file: Optional[File] = None,
    secret_name: Optional[str] = None,
) -> kubernetes.client.V1Job:
    """
    Submit a redun Task to K8S.
    """
    command = get_oneshot_command(
        scratch_prefix,
        job,
        a_task,
        args,
        kwargs,
        job_options=job_options,
        code_file=code_file,
        array_uuid=array_uuid,
    )

    if array_uuid:
        job_hash = array_uuid
    else:
        assert job.eval_hash
        job_hash = job.eval_hash

    # Submit to K8S
    job_name = get_k8s_job_name(
        job_options.get("job_name_prefix", "k8s-job"),
        job_hash,
        array=bool(array_size),
    )
    result = k8s_submit(
        k8s_client,
        command,
        image=image,
        namespace=namespace,
        job_name=job_name,
        array_size=array_size,
        secret_name=secret_name,
        **get_k8s_job_options(job_options),
    )

    return result


def submit_command(
    k8s_client: k8s_utils.K8SClient,
    image: str,
    namespace: str,
    scratch_prefix: str,
    job: Job,
    command: str,
    job_options: dict = {},
    secret_name: Optional[str] = None,
) -> kubernetes.client.V1Job:
    """
    Submit a shell command to K8S
    """
    shell_command = get_script_task_command(
        scratch_prefix,
        job,
        command,
        exit_command="exit 1",
    )

    # Submit to K8S.
    assert job.eval_hash
    job_name = get_k8s_job_name(job_options.get("job_name_prefix", "k8s-job"), job.eval_hash)
    return k8s_submit(
        k8s_client,
        shell_command,
        image=image,
        namespace=namespace,
        job_name=job_name,
        secret_name=secret_name,
        **get_k8s_job_options(job_options),
    )


def k8s_paginate(request_func: Callable) -> Iterator:
    """
    Yields items from a paginated k8s API endpoint.
    """
    token: Optional[str] = None
    while True:
        response = request_func(token)
        if response.items:
            yield from response.items
        if response.metadata._continue:
            token = response.metadata._continue
        else:
            break


def k8s_list_jobs(k8s_client: k8s_utils.K8SClient) -> Iterator[V1Job]:
    """
    Returns active K8S Jobs.
    """
    yield from k8s_paginate(
        lambda _continue: k8s_client.batch.list_job_for_all_namespaces(
            watch=False, _continue=_continue
        )
    )


def k8s_describe_jobs(
    k8s_client: k8s_utils.K8SClient, job_names: List[str], namespace: str
) -> List[kubernetes.client.V1Job]:
    """
    Returns K8S Job descriptions.
    """
    batch_api = k8s_client.batch
    jobs = []
    for job_name in job_names:
        try:
            jobs.append(batch_api.read_namespaced_job(job_name, namespace=namespace))
        except ApiException:
            # TODO: Decide what to do.
            raise
    return jobs


def get_pod_logs(
    k8s_client: k8s_utils.K8SClient, pod: kubernetes.client.V1Pod, max_lines: Optional[int] = None
) -> List[str]:
    """
    Returns the logs of a K8S pod.
    """
    kwargs = {}
    if max_lines is not None:
        kwargs["tail_lines"] = max_lines

    log_response = k8s_client.core.read_namespaced_pod_log(
        pod.metadata.name, namespace=pod.metadata.namespace, timestamps=True, **kwargs
    )
    lines = log_response.split("\n")

    # Use latest container status.
    state = pod.status.container_statuses[-1].state.terminated
    if state and state.exit_code != 0:
        lines.append(f"Exit {state.exit_code} ({state.reason}): {state.message}")

    return lines


def parse_pod_logs(
    k8s_client: k8s_utils.K8SClient, pod: kubernetes.client.V1Pod, max_lines: int = 1000
) -> Iterator[str]:
    """
    Iterates through most recent logs of an K8S Job.
    """
    lines = get_pod_logs(k8s_client, pod, max_lines=max_lines)
    if len(lines) < max_lines:
        yield "\n*** Earlier logs are truncated ***\n"
    yield from lines


def get_k8s_job_pods(core_api: CoreV1Api, job_name: str) -> Iterator[V1Pod]:
    """
    Iterates the pods for a k8s job.
    """
    yield from k8s_paginate(
        lambda _continue: core_api.list_pod_for_all_namespaces(
            watch=False,
            label_selector=f"job-name={job_name}",
            _continue=_continue,
        )
    )


class K8SError(Exception):
    """K8S-specific exception raised when a k8s job fails."""

    pass


class K8STimeoutExceeded(Exception):
    """
    Custom exception to raise when K8S jobs are killed due to timeout.
    """

    pass


@register_executor("k8s")
class K8SExecutor(Executor):
    """Implementation of K8SExecutor.
    This class adapts redun jobs to the k8s API.  It uses k8s jobs
    https://kubernetes.io/docs/concepts/workloads/controllers/job/ to
    encapsulate the submissions. For workflows with many tasks, it groups
    multiple tasks into a single k8s job
    (https://kubernetes.io/docs/tasks/job/indexed-parallel-processing-static/).
    to avoid overloading the k8s scheduler.
    and it lacks a debug mode (use minikube or another k8s tool for debug).

    Here are the details of the implementation:
    - The implementation is very similar to AWSBatchExecutor, with some minor
    differences to adapt to the k8s API.  The primary difference is that array
    jobs are handled differently
    - There is no debug mode; if you want to debug K8SExecutor-based workflows,
    use minikube https://minikube.sigs.k8s.io/ to test your workflows locally.
    Minikube is powerful enough to run many workflows on a single local machine.
    - K8SExecutor submits redun tasks (shell scripts and code) as K8S Jobs which
    execute the tasks in K8S Pods (containers).  K8S Jobs wrap pods with
    additional functionality, such as retry and parallel processing.
    - Like the AWSBatchExecutor, this implementation can batch up multiple tasks
    into a single Job.  This is more efficient for the k8s scheduler, if you are
    running more than about 100 tasks.
    - There are a few situations where this implementation can fail to return
    useful log messages from k8s jobs and pods.
    """

    def __init__(self, name: str, scheduler: Optional[Scheduler] = None, config=None):
        super().__init__(name, scheduler=scheduler)
        if config is None:
            raise ValueError("K8SExecutor requires config.")

        # Required config.
        self.image = config["image"]
        self.namespace = config.get("namespace", "default")
        self.scratch_prefix = config["scratch"]

        # Optional config.
        self.code_package = parse_code_package_config(config)
        self.secret_name: Optional[str] = config.get("secret_name")
        self.use_import_aws_secrets: bool = config.getboolean("import_aws_secrets", fallback=True)
        self.secret_env_vars = config.get("secret_env_vars", fallback="").strip().split()

        self.code_file: Optional[File] = None

        # Default task options.
        self.default_task_options = {
            "vcpus": config.getint("vcpus", 1),
            "memory": config.getfloat("memory", 4),
            "retries": config.getint("retries", 1),
            "service_account_name": config.get("service_account_name", "default"),
            "job_name_prefix": config.get("job_name_prefix", k8s_utils.DEFAULT_JOB_PREFIX),
        }
        if config.get("annotations"):
            self.default_task_options["annotations"] = json.loads(config.get("annotations"))
        if config.get("k8s_labels"):
            self.default_task_options["k8s_labels"] = json.loads(config.get("k8s_labels"))
        self.use_default_k8s_labels = config.getboolean("default_k8s_labels", True)

        self.is_running = False
        self._k8s_client = k8s_utils.K8SClient()
        # We use an OrderedDict in order to retain submission order.
        self.pending_k8s_jobs: Dict[str, Union[Job, Dict[int, Job]]] = OrderedDict()
        self.preexisting_k8s_jobs: Dict[
            str, Union[str, Tuple[str, str, int, str]]
        ] = {}  # Job hash -> Job ID

        self.interval = config.getfloat("job_monitor_interval", 5.0)

        min_array_size = config.getint("min_array_size", 5)
        max_array_size = config.getint("max_array_size", 1000)

        major, minor = self._k8s_client.version()
        if major == 1 and minor < 21:
            # Versions prior to 1.21 didn't support indexed jobs
            # (https://kubernetes.io/docs/tasks/job/indexed-parallel-processing-static/)
            logger.warn(
                "kubernetes server version is too old for indexed k8s jobs, "
                "defaulting to individual redun jobs"
            )
            min_array_size = 0
            max_array_size = 0

        self.arrayer = JobArrayer(
            self._submit_jobs,
            self._on_error,
            submit_interval=self.interval,
            stale_time=config.getfloat("job_stale_time", 3.0),
            min_array_size=min_array_size,
            max_array_size=max_array_size,
        )

    def _submit_jobs(self, jobs: List[Job]) -> None:
        """
        Callback for JobArrayer to return arrays of Jobs.
        """
        if len(jobs) == 1:
            # Singleton jobs are given as single element lists.
            self._submit_single_job(jobs[0])
        else:
            self._submit_array_job(jobs)

    def _on_error(self, error: Exception) -> None:
        """
        Callback for JobArrayer to report scheduler-level errors.
        """
        assert self._scheduler
        self._scheduler.reject_job(None, error)

    def gather_inflight_jobs(self) -> None:
        """Collect existing k8s jobs and match them up to redun jobs"""
        running_arrays: Dict[str, List[Tuple[str, str, int]]] = defaultdict(list)

        # We don't currently filter on "inflight" jobs, but that's what the
        # aws_batch implementation does.
        jobs = self.get_jobs()

        # Get all running jobs by name.
        for job in jobs:
            job_name = job.metadata.name
            if not is_array_job_name(job_name):
                job_hash = get_hash_from_job_name(job_name)
                if job_hash:
                    self.preexisting_k8s_jobs[job_hash] = job.metadata.uid
                continue

            # Get all child pods of running array jobs for reuniting.
            child_pods = get_k8s_job_pods(self._k8s_client.core, job.metadata.name)
            running_arrays[job_name].extend(
                (
                    child_pod.metadata.name,
                    child_pod.metadata.uid,
                    int(
                        child_pod.metadata.annotations["batch.kubernetes.io/job-completion-index"]
                    ),
                )
                for child_pod in child_pods
            )

        # Match up running array jobs with consistent redun job naming scheme.
        for array_name, child_pod_indices in running_arrays.items():
            # Get path to array file directory on S3 from array job name.
            parent_hash = get_hash_from_job_name(array_name)
            if not parent_hash:
                continue
            eval_file = File(
                get_array_scratch_file(
                    self.scratch_prefix,
                    parent_hash,
                    SCRATCH_HASHES,
                )
            )
            if not eval_file.exists():
                # Eval file does not exist, so we cannot reunite with this array
                # job.
                continue

            # Get eval_hash for all jobs that were part of the array
            eval_hashes = cast(str, eval_file.read("r")).splitlines()

            # Now match up indices to eval hashes to populate pending jobs by
            # name.
            for (
                child_job_name,
                child_job_id,
                child_job_index,
            ) in child_pod_indices:
                job_hash = eval_hashes[child_job_index]
                self.preexisting_k8s_jobs[job_hash] = (
                    child_job_id,
                    child_job_name,
                    child_job_index,
                    parent_hash,
                )

    def _setup_secrets(self) -> None:
        """
        Setup secrets for k8s jobs.
        """
        if not self.secret_name:
            # If secret name is not given, do not setup secrets.
            return

        # Determine env vars to import through a k8s secret.
        env_vars = {}

        # Import AWS secrets.
        if self.use_import_aws_secrets:
            env_vars.update(aws_utils.get_aws_env_vars())

        # Import local env vars into secret.
        for env_var in self.secret_env_vars:
            value = os.environ.get(env_var)
            if value:
                env_vars[env_var] = value

        k8s_utils.create_k8s_secret(self._k8s_client, self.secret_name, self.namespace, env_vars)

    def _start(self) -> None:
        """
        Start monitoring thread.
        """
        if self.is_running:
            return

        self.is_running = True

        # Ensure k8s namespace exists.
        k8s_utils.create_namespace(self._k8s_client, self.namespace)

        self._setup_secrets()

        self._thread = threading.Thread(target=self._monitor, daemon=False)
        self._thread.start()

    def stop(self) -> None:
        """
        Stop Executor and monitoring thread.
        """
        self.arrayer.stop()
        self.is_running = False

    def _monitor(self) -> None:
        """
        Thread for monitoring running K8S jobs.
        """
        assert self._scheduler

        try:
            while self.is_running and (self.pending_k8s_jobs or self.arrayer.num_pending):
                self.log(
                    f"Preparing {self.arrayer.num_pending} job(s) for Job Arrays.",
                    level=logging.DEBUG,
                )
                self.log(
                    f"Waiting on {len(self.pending_k8s_jobs)} K8S job(s): "
                    + " ".join(sorted(self.pending_k8s_jobs.keys())),
                    level=logging.DEBUG,
                )

                # Copy pending_k8s_jobs.keys() since it can change due to new
                # submissions.
                pending_jobs = list(self.pending_k8s_jobs.keys())
                jobs = k8s_describe_jobs(self._k8s_client, pending_jobs, self.namespace)
                # changing this (IE, removing iter_k8s_job_status) breaks
                # inflight test jobs =
                for job in jobs:
                    self._process_k8s_job_status(job)
                time.sleep(self.interval)

        except Exception as error:
            # Since we run this is method at the top-level of a thread, we need
            # to catch all exceptions so we can properly report them to the
            # scheduler.
            self.log("_monitor got exception", level=logging.INFO)
            self._scheduler.reject_job(None, error)

        self.log("Shutting down executor...", level=logging.DEBUG)
        self.stop()

    def _process_pod_status(self, pod: V1Pod) -> Tuple[Optional[str], Optional[str]]:
        """
        Determine if there was a failure at the pod-level.

        Returns (job_status, status_reason)
        """

        # Get container states for most recent pod.
        # TODO: check sorting.
        if not pod.status.container_statuses:
            # No container yet, so do nothing.
            return None, None
        state = pod.status.container_statuses[0].state

        # Detect-specific pod-level errors that aren't caught at the job-level.
        ERROR_STATES = {"ErrImagePull", "FailedScheduling"}
        if state.waiting and state.waiting.reason in ERROR_STATES:
            return FAILED, f"{state.waiting.reason}: {state.waiting.message}"

        elif state.terminated:
            if state.terminated.exit_code == 0:
                return SUCCEEDED, None
            else:
                return FAILED, f"{state.terminated.reason}: {state.terminated.message}"

        return None, None

    def _process_single_k8s_job_status(self, job: V1Job) -> Tuple[Optional[str], Optional[str]]:
        """
        Determine status of a non-array k8s job.
        """
        if job.status.succeeded is not None and job.status.succeeded > 0:
            return SUCCEEDED, None

        elif job.status.failed is not None and job.status.failed > 0:
            try:
                status_reason = job.status.conditions[0].message
            except (KeyError, IndexError, TypeError):
                status_reason = "Job failed."

            return FAILED, f"Error: {status_reason}"

        elif job.status.conditions is not None:
            # Handle a special case where a job times out, but isn't counted
            # as a 'failure' in job.status.failed.  In this case we want to
            # reject the job early and skip reading logs from the pod
            # because the pod was already cleaned up.
            if job.status.conditions[0].type == "Failed":
                return FAILED, "Failed: Timeout exceeded."
            else:
                return None, None
        else:
            return None, None

    def _process_redun_job(
        self,
        job: Job,
        pod: V1Pod,
        job_status: str,
        status_reason: Optional[str],
        k8s_labels: List[Tuple[str, str]] = [],
    ) -> None:
        """
        Complete a redun job (done or reject).
        """
        assert self._scheduler

        if job_status == SUCCEEDED:
            # Assume a recently completed job has valid results.
            result, exists = parse_job_result(self.scratch_prefix, job)
            if exists:
                self._scheduler.done_job(job, result, job_tags=k8s_labels)
            else:
                # This can happen if job ended in an inconsistent state.
                self._scheduler.reject_job(
                    job,
                    FileNotFoundError(
                        get_job_scratch_file(
                            self.scratch_prefix,
                            job,
                            SCRATCH_OUTPUT,
                        )
                    ),
                    job_tags=k8s_labels,
                )

        elif job_status == FAILED:
            error, error_traceback = parse_job_error(self.scratch_prefix, job)
            logs = [f"*** Logs for K8S pod {pod.metadata.name}:\n"]

            # TODO: Consider displaying events in the logs since this can have
            # helpful info as well.
            # core_api = k8s_utils.get_k8s_core_client()
            # events = core_api.list_namespaced_event(
            #     pod.metadata.namespace, field_selector=f"involvedObject.name={pod.metadata.name}"
            # )

            # Job failed here, but without an exit code
            # Detect deadline exceeded here and raise exception.
            if status_reason:
                logs.append(f"statusReason: {status_reason}\n")
            try:
                logs.extend(parse_pod_logs(self._k8s_client, pod))
            except Exception as e:
                logs.append(f"Failed to parse task logs for redun job {job.id}: {e}")

            error_traceback.logs = logs
            self._scheduler.reject_job(
                job,
                error,
                error_traceback=error_traceback,
                job_tags=k8s_labels,
            )

        else:
            raise AssertionError(f"Unexpected job_status: {job_status}")

    def _process_k8s_job_status(self, job: kubernetes.client.V1Job) -> None:
        """
        Process K8S job statuses.
        """
        if job.spec.parallelism is None or job.spec.parallelism == 1:
            # Check status of single k8s job.
            pods = list(get_k8s_job_pods(self._k8s_client.core, job.metadata.name))

            if not pods:
                # No pods yet, do nothing.
                return

            # Use most recent pod. TODO: double check sorting.
            pod = pods[0]

            job_status, status_reason = self._process_single_k8s_job_status(job)
            if not job_status:
                # If job does not have status, fallback to pod.
                job_status, status_reason = self._process_pod_status(pod)
                if not job_status:
                    # Job is still running.
                    return

            # Determine redun Job and job_labels.
            redun_job: Job = cast(Job, self.pending_k8s_jobs.pop(job.metadata.name))

            k8s_labels = [("k8s_job", job.metadata.uid)]
            self._process_redun_job(redun_job, pod, job_status, status_reason, k8s_labels)

            # Clean up k8s job immediately.
            try:
                k8s_utils.delete_job(self._k8s_client, job.metadata.name, self.namespace)
            except kubernetes.client.exceptions.ApiException as e:
                self.log(
                    f"Failed to delete k8s job {job.metadata.name}: {e}",
                    level=logging.WARN,
                )
        else:
            # Check status of array k8s job.
            k8s_pods = get_k8s_job_pods(self._k8s_client.core, job.metadata.name)
            redun_jobs = cast(Dict[int, Job], self.pending_k8s_jobs[job.metadata.name])

            for pod in k8s_pods:
                job_status, status_reason = self._process_pod_status(pod)
                if not job_status:
                    # Job is still running.
                    return

                if "batch.kubernetes.io/job-completion-index" not in pod.metadata.annotations:
                    self.log(
                        f"Pod {pod.metadata.name} is missing job-completion-index:",
                        pod.metadata.annotations,
                    )
                    continue

                index = int(pod.metadata.annotations["batch.kubernetes.io/job-completion-index"])
                redun_job2: Optional[Job] = redun_jobs.pop(index, None)
                if redun_job2:
                    k8s_labels = [("k8s_job", job.metadata.uid)]
                    self._process_redun_job(redun_job2, pod, job_status, status_reason, k8s_labels)

                # When the last pod finishes, clean up the k8s job.
                if len(redun_jobs) == 0:
                    try:
                        k8s_utils.delete_job(
                            self._k8s_client, job.metadata.name, job.metadata.namespace
                        )
                    except kubernetes.client.exceptions.ApiException as e:
                        self.log(
                            f"Failed to delete k8s job {job.metadata.name}: {e}",
                            level=logging.WARN,
                        )
                    del self.pending_k8s_jobs[job.metadata.name]

    def _get_job_options(self, job: Job) -> dict:
        """
        Determine the task options for a job.

        Task options can be specified at the job-level have precedence over the
        executor-level (within `redun.ini`):
        """
        assert job.task

        job_options = job.get_options()

        task_options = {
            **self.default_task_options,
            **job_options,
        }

        # Add default k8s labels to the job.
        if self.use_default_k8s_labels:
            execution = job.execution
            project = (
                execution.job.task.namespace
                if execution and execution.job and execution.job.task
                else ""
            )
            default_labels = {
                "redun_job_id": job.id,
                "redun_task_name": job.task.fullname,
                "redun_execution_id": execution.id if execution else "",
                "redun_project": project,
            }
        else:
            default_labels = {}

        # Merge k8s_labels if needed.
        k8s_labels = {
            **self.default_task_options.get("k8s_labels", {}),
            **default_labels,
            **job_options.get("k8s_labels", {}),
        }
        if k8s_labels:
            task_options["k8s_labels"] = k8s_labels

        return task_options

    def _submit(self, job: Job) -> None:
        """
        Submit Job to executor.
        """
        assert self._scheduler
        assert job.args
        args, kargs = job.args

        # If this is the first submission gather inflight jobs. We also check
        # is_running here as a way of determining whether this is the first
        # submission or not. If we are already running, then we know we have
        # already had jobs submitted and done the inflight check so no reason to
        # do that again here.
        if not self.is_running:
            # Precompute existing inflight jobs for job reuniting.
            self.gather_inflight_jobs()
        # Package code if necessary and we have not already done so. If
        # code_package is False, then we can skip this step. Additionally, if we
        # have already packaged and set code_file, then we do not need to
        # repackage.
        if self.code_package is not False and self.code_file is None:
            code_package = self.code_package or {}
            assert isinstance(code_package, dict)
            self.code_file = package_code(self.scratch_prefix, code_package)

        job_dir = get_job_scratch_dir(self.scratch_prefix, job)
        job_type = "K8S job"

        # Determine job options.
        task_options = self._get_job_options(job)
        cache_scope = CacheScope(task_options.get("cache_scope", CacheScope.BACKEND))

        # Determine if we can reunite with a previous K8S output or job.
        k8s_job_id: Optional[Union[str, Tuple[str, str, int, str]]] = None
        if cache_scope == CacheScope.BACKEND and job.eval_hash in self.preexisting_k8s_jobs:
            k8s_job_id = self.preexisting_k8s_jobs.pop(job.eval_hash)

            # Handle both single and array jobs.
            if isinstance(k8s_job_id, str):
                # Single job case Make sure k8s API still has a status on this
                # job.
                job_name = task_options["job_name_prefix"] + "-" + job.eval_hash
                existing_jobs = k8s_describe_jobs(
                    self._k8s_client, [job_name], namespace=self.namespace
                )
                existing_job = existing_jobs[0]  # should be index
                # Reunite with inflight k8s job, if present.
                if existing_job:
                    k8s_job_id = existing_job.metadata.uid
                    self.log(
                        "reunite redun job {redun_job} with {job_type} {k8s_job_id}:\n"
                        "  scratch_path = {job_dir}".format(
                            redun_job=job.id,
                            job_type=job_type,
                            k8s_job_id=k8s_job_id,
                            job_dir=job_dir,
                        )
                    )
                    assert k8s_job_id
                    self.pending_k8s_jobs[job_name] = job
                else:
                    k8s_job_id = None

            elif isinstance(k8s_job_id, tuple):
                # Array job case
                _, _, child_job_index, parent_hash = cast(Tuple[str, str, int, str], k8s_job_id)
                job_name = f"{task_options['job_name_prefix']}-{parent_hash}-array"
                existing_jobs = k8s_describe_jobs(
                    self._k8s_client, [job_name], namespace=self.namespace
                )
                # Reunite with inflight k8s job, if present.
                if existing_jobs:
                    existing_job = existing_jobs[0]
                    k8s_job_id = existing_job.metadata.uid
                    self.log(
                        "reunite redun job {redun_job} with {job_type} {k8s_job_id}:\n"
                        "  scratch_path = {job_dir}".format(
                            redun_job=job.id,
                            job_type=job_type,
                            k8s_job_id=k8s_job_id,
                            job_dir=job_dir,
                        )
                    )
                    assert k8s_job_id
                    if job_name not in self.pending_k8s_jobs:
                        self.pending_k8s_jobs[job_name] = cast(Dict[int, Job], {})
                    cast(Dict[int, Job], self.pending_k8s_jobs[job_name])[child_job_index] = job
                else:
                    k8s_job_id = None
            else:
                raise AssertionError(f"Unknown k8s_job_id type: {type(k8s_job_id)}")

        # Job arrayer will handle actual submission after bunching to an array
        # job, if necessary.
        if k8s_job_id is None:
            self.arrayer.add_job(job)

        self._start()

    def _submit_array_job(self, jobs: List[Job]) -> str:
        """Submits an array job, returning job name uuid"""
        array_size = len(jobs)
        all_args = []
        all_kwargs = []
        for job in jobs:
            assert job.args
            all_args.append(job.args[0])
            all_kwargs.append(job.args[1])

        # All jobs identical so just grab the first one
        job = jobs[0]
        assert job.task
        if job.task.script:
            raise NotImplementedError("Array jobs not supported for scripts")

        task_options = self._get_job_options(job)
        image = task_options.pop("image", self.image)
        namespace = task_options.pop("namespace", self.namespace)
        # Generate a unique name for job with no '-' to simplify job name
        # parsing.
        array_uuid = str(uuid.uuid4()).replace("-", "")

        job_type = "K8S job"

        # Setup input, output and error path files. Input file is a pickled list
        # of args, and kwargs, for each child job.
        input_file = get_array_scratch_file(self.scratch_prefix, array_uuid, SCRATCH_INPUT)
        with File(input_file).open("wb") as out:
            pickle_dump([all_args, all_kwargs], out)

        # Output file is a plaintext list of output paths, for each child job.
        output_file = get_array_scratch_file(self.scratch_prefix, array_uuid, SCRATCH_OUTPUT)
        output_paths = [
            get_job_scratch_file(self.scratch_prefix, job, SCRATCH_OUTPUT) for job in jobs
        ]
        with File(output_file).open("w") as ofile:
            json.dump(output_paths, ofile)

        # Error file is a plaintext list of error paths, one for each child job.
        error_file = get_array_scratch_file(self.scratch_prefix, array_uuid, SCRATCH_ERROR)
        error_paths = [
            get_job_scratch_file(self.scratch_prefix, job, SCRATCH_ERROR) for job in jobs
        ]
        with File(error_file).open("w") as efile:
            json.dump(error_paths, efile)

        # Eval hash file is plaintext hashes of child jobs for matching for job
        # reuniting.
        eval_file = get_array_scratch_file(self.scratch_prefix, array_uuid, SCRATCH_HASHES)
        with File(eval_file).open("w") as eval_f:
            eval_f.write("\n".join([job.eval_hash for job in jobs]))  # type: ignore

        k8s_resp = submit_task(
            self._k8s_client,
            image,
            namespace,
            self.scratch_prefix,
            job,
            job.task,
            job_options=task_options,
            code_file=self.code_file,
            array_uuid=array_uuid,
            array_size=array_size,
            secret_name=self.secret_name,
        )

        # Add entire array to array jobs, and all jobs in array to pending jobs.
        array_job_name = k8s_resp.metadata.name
        self.pending_k8s_jobs[array_job_name] = {}
        for i in range(array_size):
            cast(Dict[int, Job], self.pending_k8s_jobs[array_job_name])[i] = jobs[i]
        array_job_id = k8s_resp.metadata.uid

        self.log(
            "submit {array_size} redun job(s) as {job_type} {k8s_job_id}:\n"
            "  array_job_id   = {array_job_id}\n"
            "  array_job_name = {job_name}\n"
            "  array_size     = {array_size}\n"
            "  scratch_path   = {job_dir}".format(
                array_job_id=array_job_id,
                job_type=job_type,
                array_size=array_size,
                k8s_job_id=array_job_id,
                job_dir=get_array_scratch_file(self.scratch_prefix, array_uuid, ""),
                job_name=array_job_name,
            )
        )
        return array_uuid

    def _submit_single_job(self, job: Job) -> None:
        """
        Actually submits a job.  Caching detects if it should be part of an array job.
        """
        assert job.args
        args, kwargs = job.args
        task_options = self._get_job_options(job)
        image = task_options.pop("image", self.image)
        namespace = task_options.pop("namespace", self.namespace)

        job_dir = get_job_scratch_dir(self.scratch_prefix, job)
        job_type = "K8S job"

        # Submit a new Batch job.
        if not job.task.script:
            k8s_resp = submit_task(
                self._k8s_client,
                image,
                namespace,
                self.scratch_prefix,
                job,
                job.task,
                args=args,
                kwargs=kwargs,
                job_options=task_options,
                code_file=self.code_file,
                secret_name=self.secret_name,
            )
        else:
            command = get_task_command(job.task, args, kwargs)
            k8s_resp = submit_command(
                self._k8s_client,
                image,
                namespace,
                self.scratch_prefix,
                job,
                command,
                job_options=task_options,
                secret_name=self.secret_name,
            )

        k8s_job_id = k8s_resp.metadata.uid
        job_name = k8s_resp.metadata.name
        self.log(
            "submit redun job {redun_job} as {job_type} {k8s_job_id}:\n"
            "  job_id       = {k8s_job_id}\n"
            "  job_name     = {job_name}\n"
            "  scratch_path = {job_dir}".format(
                redun_job=job.id,
                job_type=job_type,
                k8s_job_id=k8s_job_id,
                job_dir=job_dir,
                job_name=job_name,
            )
        )
        self.pending_k8s_jobs[job_name] = job

    def submit(self, job: Job) -> None:
        """
        Submit Job to executor.
        """
        return self._submit(job)

    def submit_script(self, job: Job) -> None:
        """
        Submit Job for script task to executor.
        """
        return self._submit(job)

    def get_jobs(self) -> Iterator[V1Job]:
        """
        Iterates active k8s jobs.
        """
        return k8s_list_jobs(self._k8s_client)



import os
import subprocess
import warnings
from inspect import isclass, signature

import pytest
from numpydoc.docscrape import ClassDoc, FunctionDoc

import redun
from redun.tests.utils import (
    docstring_owner_pretty_name,
    get_docstring_owners_in_module,
    import_all_modules,
)


@pytest.mark.parametrize(
    "docstring_owner",
    [
        owner
        for module in import_all_modules(redun)
        for owner in get_docstring_owners_in_module(module)
    ],
    ids=docstring_owner_pretty_name,
)
def test_docstrings(docstring_owner):
    """Check if the docstring of `docstring_owner` adheres to the numpy standard.

    Currently enforces
    - Docstring section names https://numpydoc.readthedocs.io/en/latest/format.html#sections
    - Documented `Parameters` match the defined arguments

    Notes
    -----
    This function naively assumes that if `docstring_owner` is not a class,
    then its docstring should be parsable using `FunctionDoc`.
    """
    doc_parser = ClassDoc if isclass(docstring_owner) else FunctionDoc
    # Raise warnings as errors during docstring parsing.
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        parsed_doc_string = doc_parser(docstring_owner)

    # Strip leading star(s) for before comparing.
    # https://numpydoc.readthedocs.io/en/latest/format.html#parameters
    documented_param_names = {param.name.lstrip("*") for param in parsed_doc_string["Parameters"]}
    defined_argument_names = set(signature(docstring_owner).parameters.keys()) - {"cls", "self"}
    if documented_param_names:
        assert defined_argument_names == documented_param_names, (
            f"In `{docstring_owner_pretty_name(docstring_owner)}`, "
            f"defined arguments do not match documented `Parameters`"
        )


def test_no_banned_pickle_usage():
    """
    Check that dumping of pickled content uses our internal functions, not built-in pickle.

    Python 3.8 changed the default protocol for pickle. To avoid any inconsistencies in the
    output of dumped pickle content across different python versions, we want to use our own
    functions for this as they pin the protocol and will be consistent across all python versions.
    This is important because dumping of pickled content is used in hashing.
    """
    completed_process = subprocess.run(
        [
            "git",
            "grep",
            "-n",
            "-e",
            "pickle\\.dump",
            "-e",
            "from pickle import dump",
            "--",
            "redun/*.py",
        ],
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )

    # Make sure we exited with 0 exit code otherwise grep failed and there's no point going on.
    completed_process.check_returncode()

    # Bless the imports in redun.utils via their named imports and this file.
    def _is_allowed_occurrence(line):
        return (
            "allowed_dump_func" in line
            or "allowed_dumps_func" in line
            or f"{os.path.basename(__file__)}:" in line
        )

    banned_calls = [
        line for line in completed_process.stdout.splitlines() if not _is_allowed_occurrence(line)
    ]

    # We use os.linesep below because f-strings do not allow backslashes so we can't use \n.
    assert not banned_calls, (
        f"Found {len(banned_calls)} banned usages. Use redun.utils.pickle_dump(s) instead: "
        f"{os.linesep}{os.linesep.join(banned_calls)}"
    )



import os
import threading
from typing import Dict, Iterator, List, NamedTuple, Optional, Tuple
from urllib.error import URLError
from urllib.request import urlopen

import boto3

from redun.file import File

# Constants.
DEFAULT_AWS_REGION = "us-west-2"

# Cache for AWS Clients.
_boto_clients: Dict[Tuple[int, str, str], boto3.Session] = {}


class JobStatus(NamedTuple):
    all: List[str]
    pending: List[str]
    inflight: List[str]
    success: List[str]
    failure: List[str]
    stopped: List[str]
    timeout: List[str]


def get_aws_client(service: str, aws_region: str = DEFAULT_AWS_REGION) -> boto3.Session:
    """
    Get an AWS Client with caching.
    """
    cache_key = (threading.get_ident(), service, aws_region)
    client = _boto_clients.get(cache_key)
    if not client:
        # Boto is not thread safe, so we create a client per thread using
        # `threading.get_ident()` as part of our cache key.
        # We need to create the client using a session to avoid all clients
        # sharing the same global session.
        # See: https://github.com/boto/boto3/issues/801#issuecomment-440942144
        session = boto3.session.Session()
        client = _boto_clients[cache_key] = session.client(service, region_name=aws_region)

    return client


def is_ec2_instance() -> bool:
    """
    Returns True if this process is running on an EC2 instance.

    We use the presence of a link-local address as a sign we are on an EC2 instance.
    """
    try:
        resp = urlopen("http://169.254.169.254/latest/meta-data/", timeout=1)
        return resp.status == 200
    except URLError:
        return False


def get_aws_env_vars() -> Dict[str, str]:
    """
    Determines the current AWS credentials.
    """
    session = boto3.Session()
    creds = session.get_credentials().get_frozen_credentials()
    cred_map = {
        "AWS_ACCESS_KEY_ID": creds.access_key,
        "AWS_SECRET_ACCESS_KEY": creds.secret_key,
        "AWS_SESSION_TOKEN": creds.token,
    }
    # Skip variables that are not defined.
    return {k: v for k, v in cred_map.items() if v}


def copy_to_s3(file_path: str, s3_scratch_dir: str) -> str:
    """
    Copies a file to the S3 scratch directory if it is not already on S3.
    Returns the path to the file on S3.
    """
    file = File(file_path)
    _, filename = os.path.split(file.path)

    s3_temp_file = File(f"{s3_scratch_dir.rstrip('/')}/{filename}")
    file.copy_to(s3_temp_file)
    return s3_temp_file.path


def get_default_region() -> str:
    """
    Returns the default AWS region.
    """
    return boto3.Session().region_name or DEFAULT_AWS_REGION


def get_aws_user(aws_region: str = DEFAULT_AWS_REGION) -> str:
    """
    Returns the current AWS user.
    """
    sts_client = get_aws_client("sts", aws_region=aws_region)
    response = sts_client.get_caller_identity()
    return response["Arn"]


def iter_log_stream(
    log_group_name: str,
    log_stream: str,
    limit: Optional[int] = None,
    reverse: bool = False,
    required: bool = True,
    aws_region: str = DEFAULT_AWS_REGION,
) -> Iterator[dict]:
    """
    Iterate through the events of logStream.
    """
    logs_client = get_aws_client("logs", aws_region=aws_region)
    try:
        response = logs_client.get_log_events(
            logGroupName=log_group_name,
            logStreamName=log_stream,
            startFromHead=not reverse,
            # boto API does not allow passing None, so we must fully exclude the parameter.
            **{"limit": limit} if limit else {},
        )
    except logs_client.exceptions.ResourceNotFoundException as error:
        if required:
            # If logs are required, raise an error.
            raise error
        else:
            return

    while True:
        events = response["events"]

        # If no events, we are at the end of the stream.
        if not events:
            break

        if reverse:
            events = reversed(events)
        yield from events

        if not reverse:
            next_token = response["nextForwardToken"]
        else:
            next_token = response["nextBackwardToken"]
        response = logs_client.get_log_events(
            logGroupName=log_group_name,
            logStreamName=log_stream,
            nextToken=next_token,
            # boto API does not allow passing None, so we must fully exclude the parameter.
            **{"limit": limit} if limit else {},
        )



"""Add Job.execution_id

Revision ID: d4af139b6f53
Revises: 71ec303c90e4
Create Date: 2021-08-01 05:30:58.615654

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d4af139b6f53"
down_revision = "71ec303c90e4"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("job", schema=None) as batch_op:
        batch_op.add_column(sa.Column("execution_id", sa.String(), nullable=True))
        batch_op.create_index(batch_op.f("ix_job_execution_id"), ["execution_id"], unique=False)
        batch_op.create_foreign_key(
            "ct_job_execution_id",
            "execution",
            ["execution_id"],
            ["id"],
            deferrable=True,
            initially="DEFERRED",
        )

    if op.get_bind().dialect.name == "postgresql":
        op.execute(
            """
            alter table execution
            alter constraint execution_job_id_fkey deferrable initially deferred;
            """
        )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("job", schema=None) as batch_op:
        batch_op.drop_constraint("ct_job_execution_id", type_="foreignkey")
        batch_op.drop_index(batch_op.f("ix_job_execution_id"))
        batch_op.drop_column("execution_id")

    # ### end Alembic commands ###



from enum import Enum

import pytest

from redun import File, Handle, Task, task
from redun.expression import Expression
from redun.task import PartialTask
from redun.utils import pickle_dumps
from redun.value import Bool, ProxyValue, Set, Value, get_type_registry


def test_unbound_method_call():
    # Define a custom type.
    class Row:
        def __init__(self, hash, x):
            self.hash = hash
            self.x = x

    # Create a stub Value class just for augmenting Row.
    class RowValue(Value):
        type = Row

        def get_hash(self, data=None):
            return self.hash

    row = Row("aaa", 10)

    # We can call unbound methods with a different type.
    assert RowValue.get_hash(row) == "aaa"


def test_registry_raw_value():
    """
    TypeRegistry should work with raw values.
    """
    registry = get_type_registry()

    assert isinstance(registry.get_value(10), ProxyValue)
    assert registry.get_type(int) is ProxyValue
    assert registry.get_type(int, use_default=False) is None
    assert registry.get_type_name(int) == "builtins.int"
    assert registry.get_hash(10) == "279a8e0bc0a09216ec58a8d6c9f4563e21067661"
    assert registry.serialize(10) == pickle_dumps(10)
    assert registry.deserialize("builtins.int", pickle_dumps(10)) == 10
    assert registry.preprocess(10, {}) == 10
    assert registry.postprocess(10, {}) == 10
    assert registry.parse_arg(int, "10") == 10


def test_registry_value():
    """
    TypeRegistry should with Values.
    """
    registry = get_type_registry()

    # Define a custom type.
    class Row(Value):
        type_name = "myapp.Row"

        def __init__(self, hash, x):
            self.hash = hash
            self.x = x

        def __eq__(self, other):
            return self.hash == other.hash and self.x == other.x

        def get_hash(self):
            return self.hash

        def serialize(self):
            return (self.hash + ":" + str(self.x)).encode("utf8")

        @classmethod
        def deserialize(cls, raw_type, data):
            hash, x = data.decode("utf").split(":")
            return Row(hash, int(x))

        def preprocess(self, preprocess_args):
            return Row(self.hash, self.x + 1)

        def postprocess(self, postprocess_args):
            return Row(self.hash, self.x - 1)

        @classmethod
        def parse_arg(cls, raw_type, arg):
            hash, x = arg.split(":")
            return Row(hash, int(x))

    row = Row("aaa", 10)
    assert registry.get_value(row) is row
    assert registry.get_type(Row) is Row
    assert registry.get_type_name(Row) == "myapp.Row"
    assert registry.get_hash(Row("aaa", 10)) == "aaa"
    assert registry.serialize(Row("aaa", 10)) == b"aaa:10"
    assert registry.deserialize("myapp.Row", b"aaa:10") == Row("aaa", 10)
    assert registry.preprocess(Row("aaa", 10), {}) == Row("aaa", 11)
    assert registry.postprocess(Row("aaa", 11), {}) == Row("aaa", 10)
    assert registry.parse_arg(Row, "aaa:10") == Row("aaa", 10)


def test_registry_proxy_value():
    """
    TypeRegistry should work with ProxyValue.
    """
    registry = get_type_registry()

    # Define a custom type.
    class Row:
        def __init__(self, hash, x):
            self.hash = hash
            self.x = x

        def __eq__(self, other):
            return self.hash == other.hash and self.x == other.x

    # Augment custom type Row with Value methods.
    class RowValue(ProxyValue):
        type_name = "myapp.Row"
        type = Row

        def get_hash(self, data=None):
            return self.instance.hash

        def serialize(self):
            return (self.instance.hash + ":" + str(self.instance.x)).encode("utf8")

        @classmethod
        def deserialize(cls, raw_type, data):
            hash, x = data.decode("utf").split(":")
            return Row(hash, int(x))

        def preprocess(self, preprocess_args):
            return Row(self.instance.hash, self.instance.x + 1)

        def postprocess(self, postprocess_args):
            return Row(self.instance.hash, self.instance.x - 1)

        @classmethod
        def parse_arg(cls, raw_type, arg):
            hash, x = arg.split(":")
            return Row(hash, int(x))

    row = Row("aaa", 10)
    assert registry.get_value(row).instance is row
    assert registry.get_type(Row) is RowValue
    assert registry.get_type_name(Row) == "myapp.Row"
    assert registry.get_hash(Row("aaa", 10)) == "aaa"
    assert registry.serialize(Row("aaa", 10)) == b"aaa:10"
    assert registry.deserialize("myapp.Row", b"aaa:10") == Row("aaa", 10)
    assert registry.preprocess(Row("aaa", 10), {}) == Row("aaa", 11)
    assert registry.postprocess(Row("aaa", 11), {}) == Row("aaa", 10)
    assert registry.parse_arg(Row, "aaa:10") == Row("aaa", 10)


def test_registry_redun_values():
    registry = get_type_registry()

    class MyValue(Value):
        def get_hash(self, data=None):
            return "bbb"

    assert registry.get_hash(File("hello.txt")) == "18fc941eb35e16331a0709710df7e8dc75ae2619"
    assert registry.get_hash(MyValue()) == "bbb"

    assert registry.get_type_name(File) == "redun.File"
    assert registry.get_type_name(MyValue) == "redun.tests.test_type_registry.MyValue"

    # Hash using Value.get_hash()
    @task()
    def task1():
        return 10

    assert registry.get_hash(task1) == "ec0f851458d39044ae657dd250fa9526209ae6df"

    # Hash using default serialization.
    assert registry.get_hash(10) == "279a8e0bc0a09216ec58a8d6c9f4563e21067661"


class DbHandle(Handle):
    # Example Handle subclass with no type or type_name class attributes.
    def __init__(self, name, uri):
        self.uri = uri


class DbHandle2(Handle):
    type_name = "myapp.DbHandle"

    def __init__(self, name, uri):
        self.uri = uri


def test_type_name():
    pairs = [
        # Builtin python types.
        (int, "builtins.int"),
        (float, "builtins.float"),
        (bool, "builtins.bool"),
        (str, "builtins.str"),
        (dict, "builtins.dict"),
        (list, "builtins.list"),
        (set, "builtins.set"),
        # Special case python type.
        (type(None), "builtins.NoneType"),
        # redun Values.
        (Expression, "redun.expression.Expression"),
        (File, "redun.File"),
        (Task, "redun.Task"),
        (PartialTask, "redun.PartialTask"),
        (Handle, "redun.Handle"),
        (DbHandle, "redun.tests.test_type_registry.DbHandle"),
        (DbHandle2, "myapp.DbHandle"),
    ]

    registry = get_type_registry()
    for value_type, type_name in pairs:
        assert registry.get_type_name(value_type) == type_name
        if value_type is bool:
            assert registry.parse_type_name(type_name) == Bool
        elif value_type is set:
            assert registry.parse_type_name(type_name) == Set
        else:
            assert registry.parse_type_name(type_name) == value_type


def test_parse_arg():
    registry = get_type_registry()

    assert registry.parse_arg(int, "10") == 10
    assert registry.parse_arg(str, "10") == "10"
    assert registry.parse_arg(bool, "True") is True
    assert registry.parse_arg(bool, "False") is False
    assert registry.parse_arg(File, "a/b/c.txt").path == "a/b/c.txt"


def test_bool():
    registry = get_type_registry()

    assert registry.parse_arg(bool, "True")
    assert not registry.parse_arg(bool, "False")

    assert registry.serialize(True) == pickle_dumps(True)
    assert registry.serialize(False) == pickle_dumps(False)

    assert registry.deserialize("builtins.bool", pickle_dumps(True))
    assert not registry.deserialize("builtins.bool", pickle_dumps(False))
    assert registry.get_hash(True) == "4ea83061d616c782dfc446a993f359ae97cac6c8"
    assert registry.get_hash(False) == "4b0801b82e09ef3af7a2be73506b07fe7a7d5924"


def test_set():
    registry = get_type_registry()

    assert registry.get_hash({1, 2, 3}) == "0141c474b340215635cbff413fd50b019f3d6244"
    assert registry.get_hash({3, 2, 1}) == "0141c474b340215635cbff413fd50b019f3d6244"
    assert registry.get_hash([1, 2, 3]) == "e7d06e9958da58a4162eff3ccbfc83f58b7fb71a"

    assert registry.deserialize("builtins.set", registry.serialize({1, 2, 3})) == {1, 2, 3}


def test_enum() -> None:
    """
    Enum arguments should be parsable from the command line.
    """

    class Color(Enum):
        red = 1
        green = 2
        blue = 3

    registry = get_type_registry()
    assert registry.parse_arg(Color, "red") == Color.red
    with pytest.raises(ValueError):
        assert registry.parse_arg(Color, "brown") == "boom"



from typing import Any, Callable, Generic, List, Optional, Sequence, TypeVar, cast

S = TypeVar("S")
T = TypeVar("T")


class Promise(Generic[T]):
    """
    A light-weight single-thread implementation of Promise.

    We specifically are careful to remove references as soon as possible to
    enable garbage collection.

    A Promise can only be resolved or rejected once. Callbacks will be called
    at most once. Callbacks can be added both before and after the resolution.
    """

    def __init__(self, func: Optional[Callable] = None):
        self.is_pending = True
        self.is_fulfilled: Optional[bool] = None
        self.is_rejected: Optional[bool] = None

        self._value: Optional[T] = None
        self._error: Optional[Exception] = None
        self._resolvers: List[Callable[[T], S]] = []
        self._rejectors: List[Callable[[Exception], S]] = []

        if func:
            try:
                func(self.do_resolve, self.do_reject)
            except Exception as error:
                self.do_reject(error)

    @property
    def value(self) -> T:
        """
        Returns final result or error.
        """
        if self.is_pending:
            raise ValueError("Promise is still pending.")
        elif self.is_fulfilled:
            return cast(T, self._value)
        else:
            # This is a backwards compatible fix.
            return cast(T, self._error)

    @property
    def error(self) -> Exception:
        """
        Returns final error.
        """
        if not self.is_rejected:
            raise ValueError("Promise is not rejected.")
        return cast(Exception, self._error)

    def do_resolve(self, result: T) -> T:
        """
        Resolve the promise to a result.
        """
        if not self.is_pending:
            # If promise is not pending, then do nothing.
            return result
        self._value = result
        self.is_fulfilled = True
        self.is_rejected = False
        self.is_pending = False
        self._notify()
        return result

    def do_reject(self, error: Exception) -> Exception:
        """
        Reject the promise with an error.
        """
        if not self.is_pending:
            # If promise is not pending, then do nothing.
            return error
        self._error = error
        self.is_fulfilled = False
        self.is_rejected = True
        self.is_pending = False
        self._notify()
        return error

    def _notify(self) -> None:
        """
        Notify all listeners of the promise.
        """
        if self.is_pending:
            # If promise is still pending, do nothing.
            return
        elif self.is_fulfilled:
            # If promise is resolved, notify new resolvers. Discard rejectors.
            resolvers = self._resolvers
            self._resolvers = []
            self._rejectors.clear()
            for resolver in resolvers:
                resolver(cast(T, self._value))
        else:
            # If promise is rejected, notify new rejectors. Discard resolvers.
            self._resolvers.clear()
            rejectors = self._rejectors
            self._rejectors = []
            for rejector in rejectors:
                rejector(cast(Exception, self._error))

    def then(
        self,
        resolver: Optional[Callable[[T], S]] = None,
        rejector: Optional[Callable[[Exception], S]] = None,
    ) -> "Promise[S]":
        """
        Register callbacks to the promise.
        """
        promise: Promise[S] = Promise()

        def wrap_callback(func):
            def wrapper(result_or_error):
                try:
                    result2 = func(result_or_error)
                    if isinstance(result2, Promise):
                        # A nested promise should propagate to the parent promise.
                        result2.then(promise.do_resolve, promise.do_reject)
                    else:
                        # Propagate the resolved value.
                        promise.do_resolve(result2)
                except Exception as error:
                    # Propagate the rejected exception.
                    promise.do_reject(error)

            return wrapper

        if resolver:
            self._resolvers.append(wrap_callback(resolver))
        else:
            # By default propagate result.
            self._resolvers.append(cast(Callable[[T], S], promise.do_resolve))
        if rejector:
            self._rejectors.append(wrap_callback(rejector))
        else:
            # By default propagate error.
            self._rejectors.append(cast(Callable[[Exception], S], promise.do_reject))

        self._notify()
        return promise

    def catch(self, rejector: Callable[[Exception], S]) -> "Promise[S]":
        """
        Register an error callback.
        """
        return self.then(None, rejector)

    @classmethod
    def all(cls, subpromises: Sequence["Promise[T]"]) -> "Promise[List[T]]":
        """
        Return a promise that waits for all subpromises to resolve.
        """
        promise: Promise[List[T]] = Promise()
        results: List[Optional[T]] = []
        num_done = 0

        def then(i: int, result: T) -> None:
            nonlocal num_done

            results[i] = result
            num_done += 1
            if num_done == len(results):
                # All subpromises are resolved now, so resolve the top-level
                # promise with the final list.
                promise.do_resolve(cast(List[T], results))

        def fail(error: Exception) -> None:
            # As soon as we get a rejection of a subpromise, we reject the
            # top-level promise.
            promise.do_reject(error)

        def make_then(i: int) -> Callable:
            # This function gives us a closure for i.
            return lambda result: then(i, result)

        results = [None] * len(subpromises)
        for i, subpromise in enumerate(subpromises):
            subpromise.then(make_then(i), fail)

        if len(results) == 0:
            # Special case for when we are given no subpromises.
            promise.do_resolve(cast(List[T], results))

        return promise


def wait_promises(subpromises: List[Promise[T]]) -> Promise[List[Promise[T]]]:
    """
    Wait for all promises to finish (either fulfill or reject).
    """
    promise: Promise[List[Promise[T]]] = Promise()
    num_done = 0

    def done(result_or_error: Any) -> None:
        nonlocal num_done
        num_done += 1
        if num_done == len(subpromises):
            promise.do_resolve(subpromises)

    for subpromise in subpromises:
        subpromise.then(done, done)

    if len(subpromises) == 0:
        # Special case for when we are given no subpromises.
        promise.do_resolve([])

    return promise
